#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import sys
import warnings
import numpy as np
import scipy.linalg as la

from adcc import lincomb

from .common import select_eigenpairs
from .LanczosIterator import LanczosIterator
from .SolverStateBase import EigenSolverStateBase
from .explicit_symmetrisation import IndexSymmetrisation


class LanczosState(EigenSolverStateBase):
    def __init__(self, iterator):
        super().__init__(iterator.matrix)
        self.n_restart = 0
        self.subspace_residual = None  # Lanczos subspace residual vector(s)
        self.subspace_vectors = None   # Current subspace vectors
        self.algorithm = "lanczos"
        self.timer = iterator.timer


def default_print(state, identifier, file=sys.stdout):
    """
    A default print function for the lanczos callback
    """
    from adcc.timings import strtime, strtime_short

    if identifier == "start" and state.n_iter == 0:
        print("Niter n_ss  max_residual  time  Ritz values",
              file=file)
    elif identifier == "next_iter":
        time_iter = state.timer.current("iteration")
        fmt = "{n_iter:3d}  {ss_size:4d}  {residual:12.5g}  {tstr:5s}"
        print(fmt.format(n_iter=state.n_iter, tstr=strtime_short(time_iter),
                         ss_size=len(state.subspace_vectors),
                         residual=np.max(state.residual_norms)),
              "", state.eigenvalues[:7], file=file)
        if hasattr(state, "subspace_orthogonality"):
            print(33 * " " + "nonorth: {:5.3g}"
                  "".format(state.subspace_orthogonality))
    elif identifier == "is_converged":
        soltime = state.timer.total("iteration")
        print("=== Converged ===", file=file)
        print("    Number of matrix applies:   ", state.n_applies, file=file)
        print("    Total solver time:          ", strtime(soltime), file=file)
    elif identifier == "restart":
        print("=== Restart ===", file=file)


def compute_true_residuals(subspace, rvals, rvecs, epair_mask):
    """
    Compute the true residuals and residual norms (and not the ones estimated from
    the Lanczos subspace).
    """
    V = subspace.subspace
    AV = subspace.matrix_product

    def form_residual(rval, rvec):
        coefficients = np.hstack((rvec, -rval * rvec))
        return lincomb(coefficients, AV + V, evaluate=True)
    residuals = [form_residual(rvals[i], rvec)
                 for i, rvec in enumerate(np.transpose(rvecs))
                 if i in epair_mask]
    eigenvectors = [lincomb(rvec, V, evaluate=True)
                    for i, rvec in enumerate(np.transpose(rvecs))
                    if i in epair_mask]
    rnorms = np.array([np.sqrt(r @ r) for r in residuals])

    return eigenvectors, residuals, rnorms


def amend_true_residuals(state, subspace, rvals, rvecs, epair_mask):
    """
    Compute the true residuals and residual norms (and not the ones estimated from
    the Lanczos subspace) and amend the `state` accordingly.
    """
    res = compute_true_residuals(subspace, rvals, rvecs, epair_mask)
    state.eigenvectors, state.residuals, state.residual_norms = res

    return state


def check_convergence(subspace, rvals, rvecs, tol):
    b = subspace.rayleigh_extension

    # Norm of the residual vector block
    norm_residual = np.sqrt(sum(subspace.residual[p] @ subspace.residual[p]
                                for p in range(subspace.n_block)))

    # Minimal tolerance for convergence criterion
    # same settings as in ARPACK are used:
    #    norm(r) * norm(b^T * rvec) <= max(mintol, tol * abs(rval)
    mintol = np.finfo(float).eps * np.max(np.abs(rvals))
    eigenpair_error = []
    eigenpair_converged = np.ones_like(rvals, dtype=bool)
    for i, rval in enumerate(rvals):
        lhs = norm_residual * np.linalg.norm(b.T @ rvecs[:, i])
        rhs = max(mintol, tol * abs(rval))
        if lhs > rhs:
            eigenpair_converged[i] = False
        if mintol < tol * abs(rval):
            eigenpair_error.append(lhs / abs(rval))
        else:
            eigenpair_error.append(lhs * tol / mintol)
    return eigenpair_converged, np.array(eigenpair_error)


def lanczos_iterations(iterator, n_ep, min_subspace, max_subspace, conv_tol=1e-9,
                       which="LA", max_iter=100, callback=None,
                       debug_checks=False, state=None):
    """Drive the Lanczos iterations

    Parameters
    ----------
    iterator : LanczosIterator
        Iterator generating the Lanczos subspace (contains matrix, guess,
        residual, Ritz pairs from restart, symmetrisation and orthogonalisation)
    n_ep : int
        Number of eigenpairs to be computed
    min_subspace : int
        Subspace size to collapse to when performing a thick restart.
    max_subspace : int
        Maximal subspace size
    conv_tol : float, optional
        Convergence tolerance on the l2 norm of residuals to consider
        them converged
    which : str, optional
        Which eigenvectors to converge to (e.g. LM, LA, SM, SA)
    max_iter : int, optional
        Maximal number of iterations
    callback : callable, optional
        Callback to run after each iteration
    debug_checks : bool, optional
        Enable some potentially costly debug checks
        (Loss of orthogonality etc.)
    """
    if callback is None:
        def callback(state, identifier):
            pass

    if state is None:
        state = LanczosState(iterator)
        callback(state, "start")
        state.timer.restart("iteration")
        n_applies_offset = 0
    else:
        n_applies_offset = state.n_applies

    for subspace in iterator:
        b = subspace.rayleigh_extension
        with state.timer.record("rayleigh_ritz"):
            rvals, rvecs = np.linalg.eigh(subspace.subspace_matrix)

        if debug_checks:
            eps = np.finfo(float).eps
            orthotol = max(conv_tol / 1000, subspace.n_problem * eps)
            orth = subspace.check_orthogonality(orthotol)
            state.subspace_orthogonality = orth

        is_rval_converged, eigenpair_error = check_convergence(subspace, rvals,
                                                               rvecs, conv_tol)

        # Update state
        state.n_iter += 1
        state.n_applies = subspace.n_applies + n_applies_offset
        state.converged = False
        state.eigenvectors = None  # Not computed in Lanczos
        state.subspace_vectors = subspace.subspace
        state.subspace_residual = subspace.residual

        epair_mask = select_eigenpairs(rvals, n_ep, which)
        state.eigenvalues = rvals[epair_mask]
        state.residual_norms = eigenpair_error[epair_mask]
        converged = np.all(is_rval_converged[epair_mask])

        callback(state, "next_iter")
        state.timer.restart("iteration")

        if converged:
            state = amend_true_residuals(state, subspace, rvals,
                                         rvecs, epair_mask)
            state.converged = True
            callback(state, "is_converged")
            state.timer.stop("iteration")
            return state

        if state.n_iter >= max_iter:
            warnings.warn(la.LinAlgWarning(
                f"Maximum number of iterations (== {max_iter}) "
                "reached in lanczos procedure."))
            state = amend_true_residuals(state, subspace, rvals,
                                         rvecs, epair_mask)
            state.timer.stop("iteration")
            state.converged = False
            return state

        if len(rvecs) + subspace.n_block > max_subspace:
            callback(state, "restart")

            epair_mask = select_eigenpairs(rvals, min_subspace, which)
            V = subspace.subspace
            vn, betan = subspace.ortho.qr(subspace.residual)

            Y = [lincomb(rvec, V, evaluate=True)
                 for i, rvec in enumerate(np.transpose(rvecs))
                 if i in epair_mask]
            Theta = rvals[epair_mask]
            Sigma = rvecs[:, epair_mask].T @ b @ betan.T

            iterator = LanczosIterator(
                iterator.matrix, vn, ritz_vectors=Y, ritz_values=Theta,
                ritz_overlaps=Sigma,
                explicit_symmetrisation=iterator.explicit_symmetrisation
            )
            state.n_restart += 1
            return lanczos_iterations(
                iterator, n_ep, min_subspace, max_subspace, conv_tol, which,
                max_iter, callback, debug_checks, state)

    state = amend_true_residuals(state, subspace, rvals, rvecs, epair_mask)
    state.timer.stop("iteration")
    state.converged = False
    warnings.warn(la.LinAlgWarning(
        "Lanczos procedure found maximal subspace possible. Iteration cannot be "
        "continued like this and will be aborted without convergence. "
        "Try a different guess."))
    return state


def lanczos(matrix, guesses, n_ep, max_subspace=None,
            conv_tol=1e-9, which="LM", max_iter=100,
            callback=None, debug_checks=False,
            explicit_symmetrisation=IndexSymmetrisation,
            min_subspace=None):
    """Lanczos eigensolver for ADC problems

    Parameters
    ----------
    matrix
        ADC matrix instance
    guesses : list
        Guess vectors (fixes also the Lanczos block size)
    n_ep : int
        Number of eigenpairs to be computed
    max_subspace : int or NoneType, optional
        Maximal subspace size
    conv_tol : float, optional
        Convergence tolerance on the l2 norm squared of residuals to consider
        them converged
    which : str, optional
        Which eigenvectors to converge to (e.g. LM, LA, SM, SA)
    max_iter : int, optional
        Maximal number of iterations
    callback : callable, optional
        Callback to run after each iteration
    debug_checks : bool, optional
        Enable some potentially costly debug checks
        (Loss of orthogonality etc.)
    explicit_symmetrisation : optional
        Explicit symmetrisation to use after orthogonalising the
        subspace vectors. Allows to correct for loss of index or spin
        symmetries during orthogonalisation (type or instance).
    min_subspace : int or NoneType, optional
        Subspace size to collapse to when performing a thick restart.
    """
    if explicit_symmetrisation is not None and \
            isinstance(explicit_symmetrisation, type):
        explicit_symmetrisation = explicit_symmetrisation(matrix)
    iterator = LanczosIterator(matrix, guesses,
                               explicit_symmetrisation=explicit_symmetrisation)

    if not isinstance(guesses, list):
        guesses = [guesses]
    if not max_subspace:
        max_subspace = max(2 * n_ep + len(guesses), 20, 8 * len(guesses))
    if not min_subspace:
        min_subspace = n_ep + 2 * len(guesses)
    if conv_tol < matrix.shape[1] * np.finfo(float).eps:
        warnings.warn(la.LinAlgWarning(
            "Convergence tolerance (== {:5.2g}) lower than "
            "estimated maximal numerical accuracy (== {:5.2g}). "
            "Convergence might be hard to achieve."
            "".format(conv_tol, matrix.shape[1] * np.finfo(float).eps)
        ))

    return lanczos_iterations(iterator, n_ep, min_subspace, max_subspace,
                              conv_tol, which, max_iter, callback, debug_checks)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from math import sqrt

from adcc import block as b
from adcc.LazyMp import LazyMp
from adcc.AdcMethod import IsrMethod
from adcc.functions import einsum, evaluate
from adcc.Intermediates import Intermediates
from adcc.AmplitudeVector import AmplitudeVector


def mtm_isr0(mp, op, intermediates):
    f1 = op.vo.transpose()
    return AmplitudeVector(ph=f1)


def mtm_isr1(mp, op, intermediates):
    ampl = mtm_isr0(mp, op, intermediates)
    f1 = - 1.0 * einsum("ijab,jb->ia", mp.t2(b.oovv), op.ov)
    return ampl + AmplitudeVector(ph=f1)


def mtm_isr2(mp, op, intermediates):
    t2 = mp.t2(b.oovv)
    p0 = mp.mp2_diffdm

    ampl = mtm_isr1(mp, op, intermediates)
    f1 = (
        + 0.5 * einsum("ijab,jkbc,ck->ia", t2, t2, op.vo)
        + 0.5 * einsum("ij,aj->ia", p0.oo, op.vo)
        - 0.5 * einsum("bi,ab->ia", op.vo, p0.vv)
        + 1.0 * einsum("ib,ab->ia", p0.ov, op.vv)
        - 1.0 * einsum("ji,ja->ia", op.oo, p0.ov)
        - 1.0 * einsum("ijab,jb->ia", mp.td2(b.oovv), op.ov)
    )
    f2 = (
        + 1.0 * einsum("ijac,bc->ijab", t2, op.vv).antisymmetrise(2, 3)
        + 1.0 * einsum("ki,jkab->ijab", op.oo, t2).antisymmetrise(0, 1)
    )
    return ampl + AmplitudeVector(ph=f1, pphh=f2)


def mtm_cvs_isr0(mp, op, intermediates):
    f1 = op.vc.transpose()
    return AmplitudeVector(ph=f1)


def mtm_cvs_isr2(mp, op, intermediates):

    ampl = mtm_cvs_isr0(mp, op, intermediates)
    cvs_p0 = mp.second_order_dm_correction(apply_cvs=True)
    f1 = (
        - 0.5 * einsum("bI,ab->Ia", op.vc, cvs_p0.vv)
        - 1.0 * einsum("jI,ja->Ia", op.oc, cvs_p0.ov)
    )
    f2 = (1 / sqrt(2)) * einsum("kI,kjab->jIab", op.oc, mp.t2(b.oovv))
    return ampl + AmplitudeVector(ph=f1, pphh=f2)


DISPATCH = {
    "isr0": mtm_isr0,
    "isr1s": mtm_isr1,  # Identical to ISR(1)
    "isr1": mtm_isr1,
    "isr2d": mtm_isr2,  # Identical to ISR(2)
    "isr2": mtm_isr2,
    "cvs-isr0": mtm_cvs_isr0,
    "cvs-isr1s": mtm_cvs_isr0,  # Identical to CVS-ISR(0)
    "cvs-isr1": mtm_cvs_isr0,  # Identical to CVS-ISR(0)
    "cvs-isr2d": mtm_cvs_isr2,  # Identical to CVS-ISR(2)
    "cvs-isr2": mtm_cvs_isr2,
}


def modified_transition_moments(method, ground_state, operator=None,
                                intermediates=None):
    """Compute the modified transition moments (MTM) for the provided
    ISR method with reference to the passed ground state.

    Parameters
    ----------
    method: adcc.IsrMethod
        Provide a method at which to compute the MTMs
    ground_state : adcc.LazyMp
        The MP ground state
    operator : adcc.OneParticleOperator or list, optional
        Only required if different operators than the standard
        electric dipole operators in the MO basis should be used.
    intermediates : adcc.Intermediates
        Intermediates from the ADC calculation to reuse

    Returns
    -------
    adcc.AmplitudeVector or list of adcc.AmplitudeVector
    """
    if not isinstance(method, IsrMethod):
        method = IsrMethod(method)
    if not isinstance(ground_state, LazyMp):
        raise TypeError("ground_state should be a LazyMp object.")
    if intermediates is None:
        intermediates = Intermediates(ground_state)

    unpack = False
    if operator is None:
        operator = ground_state.reference_state.operators.electric_dipole
    elif not isinstance(operator, (list, tuple)):
        unpack = True
        operator = [operator]
    if method.name not in DISPATCH:
        raise NotImplementedError("modified_transition_moments is not "
                                  f"implemented for {method.name}.")

    ret = [DISPATCH[method.name](ground_state, op, intermediates)
           for op in operator]
    if unpack:
        assert len(ret) == 1
        ret = ret[0]
    return evaluate(ret)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from import_data import import_data

# Gather precomputed data
data = import_data()

# Run an adc2 calculation:
singlets = adcc.adc2(data, frozen_core=1, n_singlets=5, conv_tol=1e-8)
triplets = adcc.adc2(singlets.matrix, frozen_core=1, n_triplets=5, conv_tol=1e-8)
print(singlets.describe())
print(triplets.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from .misc import cached_property
from .Tensor import Tensor
from .MoSpaces import MoSpaces
from .backends import import_scf_results
from .OperatorIntegrals import OperatorIntegrals
from .OneParticleDensity import OneParticleDensity
from .TwoParticleDensity import TwoParticleDensity
from .NParticleOperator import product_trace, OperatorSymmetry

import libadcc

__all__ = ["ReferenceState"]


class ReferenceState(libadcc.ReferenceState):
    def __init__(self, hfdata, core_orbitals=None, frozen_core=None,
                 frozen_virtual=None, symmetry_check_on_import=False,
                 import_all_below_n_orbs=10):
        """Construct a ReferenceState holding information about the employed
        SCF reference.

        The constructed object is lazy and will at construction only setup
        orbital energies and coefficients. Fock matrix blocks and
        electron-repulsion integral blocks are imported as needed.

        Orbital subspace selection: In order to specify `frozen_core`,
        `core_orbitals` and `frozen_virtual`, adcc allows a range of
        specifications including

           a. A number: Just put this number of alpha orbitals and this
              number of beta orbitals into the respective space. For frozen
              core and core orbitals these are counted from below, for
              frozen virtual orbitals, these are counted from above. If both
              frozen core and core orbitals are specified like this, the
              lowest-energy, occupied orbitals will be put into frozen core.
           b. A range: The orbital indices given by this range will be put
              into the orbital subspace.
           c. An explicit list of orbital indices to be placed into the
              subspace.
           d. A pair of (a) to (c): If the orbital selection for alpha and
              beta orbitals should differ, a pair of ranges, or a pair of
              index lists or a pair of numbers can be specified.

        Parameters
        ----------
        hfdata
            Object with Hartree-Fock data (e.g. a molsturm scf state, a pyscf
            SCF object or any class implementing the
            :py:class:`adcc.HartreeFockProvider` interface or in fact any python
            object representing a pointer to a C++ object derived off
            the :cpp:class:`adcc::HartreeFockSolution_i`.

        core_orbitals : int or list or tuple, optional
            The orbitals to be put into the core-occupied space. For ways to
            define the core orbitals see the description above.

        frozen_core : int or list or tuple, optional
            The orbitals to be put into the frozen core space. For ways to
            define the core orbitals see the description above. For an automatic
            selection of the frozen core space one may also specify
            ``frozen_core=True``.

        frozen_virtuals : int or list or tuple, optional
            The orbitals to be put into the frozen virtual space. For ways to
            define the core orbitals see the description above.

        symmetry_check_on_import : bool, optional
            Should symmetry of the imported objects be checked explicitly during
            the import process. This massively slows down the import and has a
            dramatic impact on memory usage. Thus one should enable this only
            for debugging (e.g. for testing import routines from the host
            programs). Do not enable this unless you know what you are doing.

        import_all_below_n_orbs : int, optional
            For small problem sizes lazy make less sense, since the memory
            requirement for storing the ERI tensor is neglibile and thus the
            flexiblity gained by having the full tensor in memory is
            advantageous. Below the number of orbitals specified by this
            parameter, the class will thus automatically import all ERI tensor
            and Fock matrix blocks.

        Examples
        --------
        To start a calculation with the 2 lowest alpha and beta orbitals
        in the core occupied space, construct the class as

        >>> ReferenceState(hfdata, core_orbitals=2)

        or

        >>> ReferenceState(hfdata, core_orbitals=range(2))

        or

        >>> ReferenceState(hfdata, core_orbitals=[0, 1])

        or

        >>> ReferenceState(hfdata, core_orbitals=([0, 1], [0, 1]))

        There is no restriction to choose the core occupied orbitals
        from the bottom end of the occupied orbitals. For example
        to select the 2nd and 3rd orbital setup the class as

        >>> ReferenceState(hfdata, core_orbitals=range(1, 3))

        or

        >>> ReferenceState(hfdata, core_orbitals=[1, 2])

        If different orbitals should be placed in the alpha and
        beta orbitals, this can be achievd like so

        >>> ReferenceState(hfdata, core_orbitals=([1, 2], [0, 1]))

        which would place the 2nd and 3rd alpha and the 1st and second
        beta orbital into the core space.
        """
        if not isinstance(hfdata, libadcc.HartreeFockSolution_i):
            hfdata = import_scf_results(hfdata)

        self._mospaces = MoSpaces(hfdata, frozen_core=frozen_core,
                                  frozen_virtual=frozen_virtual,
                                  core_orbitals=core_orbitals)
        super().__init__(hfdata, self._mospaces, symmetry_check_on_import)

        if import_all_below_n_orbs is not None and \
           hfdata.n_orbs < import_all_below_n_orbs:
            super().import_all()

        self.operators = OperatorIntegrals(
            hfdata.operator_integral_provider, self._mospaces,
            self.orbital_coefficients, self.orbital_coefficients_alpha,
            self.orbital_coefficients_beta, self.conv_tol
        )

        self.environment = None  # no environment attached by default
        for name in ["excitation_energy_corrections", "environment"]:
            if hasattr(hfdata, name):
                setattr(self, name, getattr(hfdata, name))

    def __getattr__(self, attr):
        from . import block as b

        if attr.startswith("f"):
            return self.fock(b.__getattr__(attr[1:]))
        else:
            return self.eri(b.__getattr__(attr))

    @property
    def mospaces(self):
        return self._mospaces

    @property
    def timer(self):
        ret = super().timer
        ret.attach(self.operators.timer)
        return ret

    @property
    def is_aufbau_occupation(self):
        """
        Returns whether the molecular orbital occupation in this reference
        is according to the Aufbau principle (lowest-energy orbitals are occupied)
        """
        eHOMO = max(max(self.orbital_energies(space).to_ndarray())
                    for space in self.mospaces.subspaces_occupied)
        eLUMO = min(min(self.orbital_energies(space).to_ndarray())
                    for space in self.mospaces.subspaces_virtual)
        return eHOMO < eLUMO

    def to_qcvars(self, properties=False, recurse=False):
        """
        Return a dictionary with property keys compatible to a Psi4 wavefunction
        or a QCEngine Atomicresults object.
        """
        qcvars = {"HF TOTAL ENERGY": self.energy_scf, }
        if properties:
            qcvars["HF DIPOLE"] = self.dipole_moment
        return qcvars

    @property
    def density(self):
        """
        Return the Hartree-Fock density in the MO basis
        """
        density = OneParticleDensity(self.mospaces,
                                     symmetry=OperatorSymmetry.HERMITIAN)
        for block in density.canonical_blocks:
            sym = libadcc.make_symmetry_operator(self.mospaces, block,
                                                 density.symmetry.to_str(), "1")
            density[block] = Tensor(sym)
        for ss in self.mospaces.subspaces_occupied:
            density[ss + ss].set_mask("ii", 1)
        density.reference_state = self
        return density

    @property
    def density_2p(self):
        """
        Return the two-particle Hartree-Fock density in the MO basis
        """
        density = TwoParticleDensity(self.mospaces,
                                     symmetry=OperatorSymmetry.HERMITIAN)
        for block in density.canonical_blocks:
            sym = libadcc.make_symmetry_operator(self.mospaces, block,
                                                 density.symmetry.to_str(), "1")
            density[block] = Tensor(sym)
        for ss in self.mospaces.subspaces_occupied:
            density[ss + ss + ss + ss].set_mask("ijij", 1)
            density[ss + ss + ss + ss].set_mask("ijji", -1)
            density[ss + ss + ss + ss].set_mask("iijj", 0)
        density.reference_state = self
        return density

    @cached_property
    def dipole_moment(self):
        """
        Return the HF dipole moment of the reference state (that is the sum of
        the electronic and the nuclear contribution.)
        """
        dipole_integrals = self.operators.electric_dipole
        return self.nuclear_dipole + np.array([product_trace(comp, self.density)
                                               for comp in dipole_integrals])

    def nuclear_quadrupole(self, gauge_origin="origin"):
        if isinstance(gauge_origin, str):
            gauge_origin = self.gauge_origin_to_xyz(gauge_origin)
        return super().nuclear_quadrupole(gauge_origin)

    @cached_property
    def ssq(self):
        """
        Return <S^2> of the HF reference state.
        """
        if self.restricted:
            raise NotImplementedError(
                "<S^2> is not implemented for restricted HF references."
            )
        ssq_1p_op = self.operators.ssq_1p
        ssq_2p_op = self.operators.ssq_2p
        ssq_1p = product_trace(ssq_1p_op, self.density)
        ssq_2p = product_trace(ssq_2p_op, self.density_2p)
        return (ssq_1p + ssq_2p)

# TODO some nice describe method



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2026 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .MoSpaces import MoSpaces, split_spaces
from .NParticleOperator import OperatorSymmetry, product_trace
from .OneParticleDensity import OneParticleDensity
from .ReferenceState import ReferenceState
from .TwoParticleDensity import TwoParticleDensity
from .functions import direct_sum, einsum, zeros_like
from .misc import cached_member_function
from .timings import Timer
from . import block as b

import libadcc

from typing import Union
import numpy as np


class GroundState:
    """
    Base class representing the ground state.
    """
    def __init__(self, hf: Union[ReferenceState, libadcc.HartreeFockSolution_i]):
        if isinstance(hf, libadcc.HartreeFockSolution_i):
            hf = ReferenceState(hf)
        if not isinstance(hf, ReferenceState):
            raise TypeError("hf needs to be a 'ReferenceState' "
                            "or a 'HartreeFockSolution_i'")
        self.reference_state: ReferenceState = hf
        self.mospaces: MoSpaces = hf.mospaces
        self.timer: Timer = Timer()
        self.has_core_occupied_space: bool = hf.has_core_occupied_space

    def td1(self, space: str) -> libadcc.Tensor:
        """First-order ground state doubles amplitudes."""
        raise NotImplementedError(
            "First-order ground state doubles amplitudes not implemented on "
            f"{self.__class__.__name__} class."
        )

    def t2(self, space: str) -> libadcc.Tensor:
        """Alias for td1 for backwards compatiblity."""
        return self.td1(space)

    @property
    def t2oo(self) -> libadcc.Tensor:
        return self.td1(b.oovv)

    @property
    def t2oc(self) -> libadcc.Tensor:
        return self.td1(b.ocvv)

    @property
    def t2cc(self) -> libadcc.Tensor:
        return self.td1(b.ccvv)

    def ts2(self, space: str, apply_cvs: bool = False) -> libadcc.Tensor:
        """Second-order ground state singles amplitudes"""
        # NOTE: In contrast to the other amplitudes, ts2 has the additional
        # apply_cvs argument, since ts2 corresponds to the OV/CV block of the
        # second order 1p density correction.
        raise NotImplementedError(
            "Second-order ground state singles amplitudes not implemented on "
            f"{self.__class__.__name__} class."
        )

    def td2(self, space: str) -> libadcc.Tensor:
        """Second-order ground state doubles amplitudes"""
        raise NotImplementedError(
            "Second-order ground state doubles amplitudes not implemented on "
            f"{self.__class__.__name__} class."
        )

    def tt2(self, space: str) -> libadcc.Tensor:
        """Second-order ground state triples amplitudes"""
        raise NotImplementedError(
            "Second-order ground state triples amplitudes not implemented on "
            f"{self.__class__.__name__} class."
        )

    def energy_correction(self, level: int = 2) -> float:
        """Obtain the ground state energy correction at a particular level"""
        raise NotImplementedError(
            "Ground state energy corrections not implemented on "
            f"{self.__class__.__name__} class."
        )

    def energy(self, level: int = 2) -> float:
        """
        Obtain the total energy (SCF energy plus all corrections)
        at a particular level of perturbation theory.
        """
        raise NotImplementedError(
            "Ground state energy not implemented on "
            f"{self.__class__.__name__} class."
        )

    def diffdm(self, level: int = 2, apply_cvs: bool = False) -> OneParticleDensity:
        """
        Return the ground state difference denstiy in the MO basis
        with all corrections up to the specified order of perturbation theory.
        """
        if level >= 0 and level < 2:
            raise ValueError(f"Difference density of order {level} vanishes.")
        elif level == 2:
            return self.second_order_dm_correction(apply_cvs=apply_cvs)
        else:
            raise NotImplementedError(
                "Only second-order density corection is implemented. "
                f"diffdm of level {level} is not available."
            )

    def density(self, level: int = 2,
                apply_cvs: bool = False) -> OneParticleDensity:
        """
        Return the ground state density in the MO basis with all corrections
        up to the specified order of perturbation theory.
        """
        if level in [0, 1]:
            return self.reference_state.density
        diffdm = self.diffdm(level, apply_cvs=apply_cvs)
        return self.reference_state.density + diffdm

    def diffdm_2p(self, level: int = 2,
                  apply_cvs: bool = False) -> TwoParticleDensity:
        """
        Return the two-particle ground state difference density in the MO basis
        with all corrections up to the specified order of perturbation theory.
        """
        if level == 0:
            raise ValueError("Zeroth-order 2-particle difference density vanishes.")
        elif level == 1:
            return self.first_order_dm_correction_2p(apply_cvs=apply_cvs)
        elif level == 2:
            return (self.first_order_dm_correction_2p(apply_cvs=apply_cvs)
                    + self.second_order_dm_correction_2p(apply_cvs=apply_cvs))
        raise NotImplementedError("Only first and second-order two-particle "
                                  "density corrections are implemented. "
                                  f"2p diffdm of level {level} is not available.")

    def density_2p(self, level: int = 2,
                   apply_cvs: bool = False) -> TwoParticleDensity:
        """
        Return the two-particle ground state density in the MO basis
        with all corrections up to the specified order of perturbation theory.
        """
        if level == 0:
            return self.reference_state.density_2p
        diffdm = self.diffdm_2p(level, apply_cvs=apply_cvs)
        return self.reference_state.density_2p + diffdm

    @cached_member_function()
    def dipole_moment(self, level: int = 2, apply_cvs: bool = False
                      ) -> np.ndarray[tuple[int], np.dtype[np.float64]]:
        """
        Return the ground state dipole moment at the specified level of
        perturbation theory.
        """
        dm = self.density(level, apply_cvs=apply_cvs)
        dipole_integrals = self.reference_state.operators.electric_dipole
        return self.reference_state.nuclear_dipole + np.array([
            product_trace(comp, dm) for comp in dipole_integrals
        ])

    @cached_member_function()
    def ssq(self, level: int = 2, apply_cvs: bool = False) -> float:
        """
        Return <S^2> of the ground state.
        """
        if self.reference_state.restricted:
            raise NotImplementedError(
                "<S^2> is not implemented for restricted HF references."
            )
        ssq_1p_op = self.reference_state.operators.ssq_1p
        ssq_2p_op = self.reference_state.operators.ssq_2p
        # the trace of the second-order (and higher) correction to the RDM1
        # is zero -> no influence on top of HF density for ground state
        ssq_1p = product_trace(ssq_1p_op, self.density(0))
        ssq_2p = product_trace(
            ssq_2p_op, self.density_2p(level, apply_cvs=apply_cvs)
        )
        return ssq_1p + ssq_2p

    @cached_member_function()
    def df(self, space: str) -> libadcc.Tensor:
        """Delta Fock matrix"""
        hf = self.reference_state
        s1, s2 = split_spaces(space)
        fC = hf.fock(s1 + s1).diagonal()
        fv = hf.fock(s2 + s2).diagonal()
        return direct_sum("-i+a->ia", fC, fv)

    @cached_member_function()
    def t2eri(self, space: str, contraction: str) -> libadcc.Tensor:
        """
        Return the T2 tensor with ERI tensor contraction intermediates.
        These are called pi1 to pi7 in libadc.
        """
        hf = self.reference_state
        key = space + contraction
        expressions = {
            # space + contraction
            b.ooov + b.vv: ('ijbc,kabc->ijka', b.ovvv),
            b.ooov + b.ov: ('ilab,lkjb->ijka', b.ooov),
            b.oovv + b.oo: ('klab,ijkl->ijab', b.oooo),
            b.oovv + b.ov: ('jkac,kbic->ijab', b.ovov),
            b.oovv + b.vv: ('ijcd,abcd->ijab', b.vvvv),
            b.ovvv + b.oo: ('jkbc,jkia->iabc', b.ooov),
            b.ovvv + b.ov: ('ijbd,jcad->iabc', b.ovvv),
        }
        if key not in expressions:
            raise NotImplementedError("t2eri intermediate not implemented "
                                      f"for space '{space}' and contraction "
                                      f"'{contraction}'.")
        contraction_str, eri_block = expressions[key]
        return einsum(contraction_str, self.t2oo, hf.eri(eri_block))

    @cached_member_function()
    def second_order_dm_correction(self, apply_cvs: bool = False
                                   ) -> OneParticleDensity:
        """
        Return the second-order contribution to the ground state
        difference density in the MO basis.
        """
        if apply_cvs and not self.has_core_occupied_space:
            raise RuntimeError("Cannot apply the CVS approximation to a "
                               "ground state build on top of a HF reference state "
                               "without a core space.")

        ret = OneParticleDensity(
            self.mospaces, symmetry=OperatorSymmetry.HERMITIAN
        )
        ret.oo = -0.5 * einsum("ikab,jkab->ij", self.t2oo, self.t2oo)
        ret.ov = self.ts2(b.ov, apply_cvs=apply_cvs)
        ret.vv = 0.5 * einsum("ijac,ijbc->ab", self.t2oo, self.t2oo)

        if self.has_core_occupied_space and not apply_cvs:
            # additional terms since we don't apply the CVS approximation
            # for the GS density. Within the CVS approximation all of the
            # following terms vanish for the MP and RE partitionings.
            # Not sure if this is true for all partitionings!
            ret.oo += -0.5 * einsum("iLab,jLab->ij", self.t2oc, self.t2oc)
            ret.vv += (
                + 0.5 * einsum("IJac,IJbc->ab", self.t2cc, self.t2cc)
                + 1.0 * einsum("kJac,kJbc->ab", self.t2oc, self.t2oc)
            )
            # compute extra CVS blocks
            ret.cc = -0.5 * (
                + einsum("kIab,kJab->IJ", self.t2oc, self.t2oc)
                + einsum('LIab,LJab->IJ', self.t2cc, self.t2cc)
            )
            ret.oc = -0.5 * (
                + einsum("kIab,kjab->jI", self.t2oc, self.t2oo)
                + einsum("ILab,jLab->jI", self.t2cc, self.t2oc)
            )
            ret.cv = self.ts2(b.cv, apply_cvs=apply_cvs)
        ret.reference_state = self.reference_state
        return ret.evaluate()

    @cached_member_function()
    def first_order_dm_correction_2p(self, apply_cvs: bool = False
                                     ) -> TwoParticleDensity:
        """
        Return the two-particle first-order difference density correction
        in the MO basis.
        """
        if self.has_core_occupied_space:
            raise NotImplementedError("First-order 2-particle DM correction not "
                                      "implemented for a ground state with "
                                      "core orbitals.")
        assert not apply_cvs  # TODO: once implemented for core orbitals
        ret = TwoParticleDensity(
            self.mospaces, symmetry=OperatorSymmetry.HERMITIAN
        )
        ret.oovv = -1.0 * self.t2oo
        ret.reference_state = self.reference_state
        return ret.evaluate()

    @cached_member_function()
    def second_order_dm_correction_2p(self, apply_cvs: bool = False
                                      ) -> TwoParticleDensity:
        """
        Return the two-particle second-order difference density correction
        in the MO basis.
        """
        if self.has_core_occupied_space:
            raise NotImplementedError("Second-order 2-particle DM correction not "
                                      "implemented for a ground state with "
                                      "core orbitals.")
        assert not apply_cvs  # TODO: once implemented for core orbitals
        hf: ReferenceState = self.reference_state
        ret = TwoParticleDensity(
            self.mospaces, symmetry=OperatorSymmetry.HERMITIAN
        )
        p0: OneParticleDensity = self.diffdm(2)

        # constuct Kronecker Delta
        d_oo = zeros_like(hf.foo)
        d_oo.set_mask("ii", 1)

        ret.oooo = (
            + 4.0 * einsum("ik,jl->ijkl", p0.oo, d_oo)
            .antisymmetrise(0, 1).antisymmetrise(2, 3)
            + 0.5 * einsum("ijab,klab->ijkl", self.t2oo, self.t2oo)
        )
        ret.ooov = (
            + 2.0 * einsum("ja,ik->ijka", p0.ov, d_oo).antisymmetrise(0, 1)
        )
        ret.oovv = (
            - 1.0 * self.td2(b.oovv)
        )
        ret.ovov = (
            + 1.0 * einsum("ab,ij->iajb", p0.vv, d_oo)
            - 1.0 * einsum("jkac,ikbc->iajb", self.t2oo, self.t2oo)
        )
        ret.vvvv = (
            + 0.5 * einsum("ijab,ijcd->abcd", self.t2oo, self.t2oo)
        )
        ret.reference_state = self.reference_state
        return ret.evaluate()

    def _to_qcvars(self, gs_type: str, properties: bool = False,
                   recurse: bool = False, maxlevel: int = 2) -> dict:
        """
        Return a dictionary with property keys compatible to a Psi4 wavefunction
        or a QCEngine Atomicresults object.
        """
        qcvars = {}
        for level in range(2, maxlevel + 1):
            try:
                mpcorr = self.energy_correction(level)
                qcvars[f"{gs_type}{level} CORRELATION ENERGY"] = mpcorr
                qcvars[f"{gs_type}{level} TOTAL ENERGY"] = self.energy(level)
            except NotImplementedError:
                pass
            except ValueError:
                pass

        if properties:
            for level in range(2, maxlevel + 1):
                try:
                    qcvars[f"{gs_type}{level} DIPOLE"] = self.dipole_moment(level)
                except NotImplementedError:
                    pass

        if recurse:
            qcvars.update(self.reference_state.to_qcvars(properties, recurse))
        return qcvars



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .MoSpaces import split_spaces
from .NParticleOperator import OperatorSymmetry

from collections.abc import Callable, Sequence
from typing import TypeVar, Any, Union


def __getattr__(attr: Sequence[str]) -> str:
    """
    Return a multi-dimensional block string like 'o1o2v1v1'
    when requesting the attribute 'ocvv'.
    """
    if any(c not in ["o", "v", "c"] for c in attr):
        raise AttributeError
    mapping = {"o": "o1", "v": "v1", "c": "o2"}
    return "".join(mapping[c] for c in attr)


T = TypeVar("T")


# adapted from
# https://github.com/sympy/sympy/blob/master/sympy/physics/secondquant.py
def _sort_anticommuting(to_sort: list[T],
                        key: Union[Callable[[T], Any], None] = None
                        ) -> tuple[list[T], int]:
    """
    Sort a list of mutually anticommuting operators into canonical order.

    All elements in ``to_sort`` are assumed to anticommute pairwise.  The
    sorting is performed via a bidirectional bubble sort.

    Parameters
    ----------
    to_sort : list[T]
        List of operators (or any comparable objects) to be sorted.
    key : callable, optional
        Optional function mapping each element of ``to_sort`` to a comparison
        key.  If ``None`` (default), the elements themselves are compared.

    Returns
    -------
    to_sort : list[T]
        The operators sorted into canonical order.
    sign : int
        The sign accumulated from the reordering, i.e. ``+1`` or ``-1``.
    """
    verified = False
    sign = 1
    rng = tuple(range(len(to_sort) - 1))
    rev = tuple(range(len(to_sort) - 3, -1, -1))

    def _identity(x):
        return x

    if key is None:
        key = _identity

    keys = list(map(key, to_sort))
    to_sort = list(to_sort)

    while not verified:
        verified = True
        for i in rng:
            left = keys[i]
            right = keys[i + 1]
            if left > right:
                verified = False
                keys[i], keys[i + 1] = right, left
                to_sort[i], to_sort[i + 1] = to_sort[i + 1], to_sort[i]
                sign *= -1
        if verified:
            break
        for i in rev:
            left = keys[i]
            right = keys[i + 1]
            if left > right:
                verified = False
                keys[i], keys[i + 1] = right, left
                to_sort[i], to_sort[i + 1] = to_sort[i + 1], to_sort[i]
                sign *= -1
    return (to_sort, sign)


def get_canonical_block(bra: str, ket: str,
                        operator_symmetry: OperatorSymmetry
                        ) -> tuple[str, int, tuple]:
    """
    Return the canonical representation of an operator block, together with the
    factor required to recover the desired block and the transpose tuple that
    encodes the transformation from the canonical form back to the requested block.

    Parameters
    ----------
    bra : str
        The bra index string of the block.
    ket : str
        The ket index string of the block.
    operator_symmetry : OperatorSymmetry
        Symmetry of the operator.

    Returns
    -------
    canonical_block : str
        The canonicalised block.
    factor : int
        The integer factor needed to recover the desired block
        from the canonical block, i.e. ``+1`` or ``-1``.
    transform : tuple
        A tuple encoding the transformation applied to obtain the desired block
        from the canonical block.
    """
    def invert_transpose_tuple(p: tuple[int, ...]) -> tuple[int, ...]:
        """
        Invert transpose tuple containing permutation that maps a block to its
        canonical form. Returns the tuple that maps from the canonical block
        back to the original block.
        """
        q = [None] * len(p)
        for i, pi in enumerate(p):
            q[pi] = i
        return tuple(q)

    factor = 1

    bra = split_spaces(bra) if bra else []
    bra_sorted, bra_factor = _sort_anticommuting(
        list(enumerate(bra)), key=lambda tpl: tpl[1]
    )
    bra_transpose = [val for val, _ in bra_sorted]
    bra_space_sorted = [val for _, val in bra_sorted]
    factor *= bra_factor

    ket = split_spaces(ket) if ket else []
    ket_sorted, ket_factor = _sort_anticommuting(
        list(enumerate(ket)), key=lambda tpl: tpl[1]
    )
    ket_transpose = [val for val, _ in ket_sorted]
    ket_transpose = tuple(x + len(bra) for x in ket_transpose)
    ket_space_sorted = [val for _, val in ket_sorted]
    factor *= ket_factor

    transpose = (*bra_transpose, *ket_transpose)
    canonical_block = "".join(bra_space_sorted) + "".join(ket_space_sorted)

    if operator_symmetry != OperatorSymmetry.NOSYMMETRY:
        if len(bra) != len(ket):
            raise ValueError("Invalid blocks strings.")

        if ket_space_sorted < bra_space_sorted:
            transpose = (*ket_transpose, *bra_transpose)
            canonical_block = "".join(ket_space_sorted) + "".join(bra_space_sorted)
            if operator_symmetry == OperatorSymmetry.ANTIHERMITIAN:
                factor *= -1
    transpose = invert_transpose_tuple(transpose)
    return (canonical_block, factor, transpose)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import os
import tempfile

from mpi4py import MPI

import adcc
import veloxchem as vlx
from veloxchem.mpitask import MpiTask

# Run SCF in VeloxChem
with tempfile.TemporaryDirectory() as tmpdir:
    infile = os.path.join(tmpdir, "vlx.in")
    outfile = os.path.join(tmpdir, "/dev/null")

    with open(infile, "w") as fp:
        fp.write("""
                 @jobs
                 task: hf
                 @end

                 @method settings
                 basis: sto-3g
                 @end

                 @molecule
                 charge: 0
                 multiplicity: 1
                 units: bohr
                 xyz:
                 O 0 0 0
                 H 0 0 1.795239827225189
                 H 1.693194615993441 0 -0.599043184453037
                 @end
                 """)
    task = MpiTask([infile, outfile], MPI.COMM_WORLD)
    scfdrv = vlx.ScfRestrictedDriver(task.mpi_comm, task.ostream)
    scfdrv.conv_thresh = 1e-8
    scfdrv.compute(task.molecule, task.ao_basis, task.min_basis)
    scfdrv.task = task

# Run an adc2 calculation:
state = adcc.adc2(scfdrv, n_singlets=5)
print(state.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import os
import adcc
import tempfile

from mpi4py import MPI
from matplotlib import pyplot as plt

import veloxchem as vlx
from veloxchem.mpitask import MpiTask

# Run SCF in VeloxChem
with tempfile.TemporaryDirectory() as tmpdir:
    infile = os.path.join(tmpdir, "vlx.in")
    outfile = os.path.join(tmpdir, "/dev/null")

    with open(infile, "w") as fp:
        fp.write("""
                 @jobs
                 task: hf
                 @end

                 @method settings
                 basis: cc-pvdz
                 @end

                 @molecule
                 charge: 0
                 multiplicity: 1
                 units: bohr
                 xyz:
                 O 0 0 0
                 H 0 0 1.795239827225189
                 H 1.693194615993441 0 -0.599043184453037
                 @end
                 """)
    task = MpiTask([infile, outfile], MPI.COMM_WORLD)
    scfdrv = vlx.ScfRestrictedDriver(task.mpi_comm, task.ostream)
    scfdrv.conv_thresh = 1e-9
    scfdrv.compute(task.molecule, task.ao_basis, task.min_basis)
    scfdrv.task = task

print(adcc.banner())

# Run an adc2 calculation:
state = adcc.adc2(scfdrv, n_singlets=7, conv_tol=1e-8)

print()
print("  st  ex.ene. (au)         f     transition dipole moment (au)"
      "        state dip (au)")
for i, val in enumerate(state.excitation_energy):
    fmt = "{0:2d}  {1:12.8g} {2:9.3g}   [{3:9.3g}, {4:9.3g}, {5:9.3g}]"
    fmt += "   [{6:9.3g}, {7:9.3g}, {8:9.3g}]"
    print(state.kind[0], fmt.format(i, val, state.oscillator_strength[i],
                                    *state.transition_dipole_moment[i],
                                    *state.state_dipole_moment[i]))

# Plot a spectrum
state.plot_spectrum()
plt.savefig("veloxchem_ccpvdz_adc2_spectrum.pdf")
plt.show()

print()
print(state.timer.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from math import sqrt
from collections import namedtuple

from adcc import block as b
from adcc.functions import direct_sum, einsum, zeros_like
from adcc.Intermediates import Intermediates, register_as_intermediate
from adcc.AmplitudeVector import AmplitudeVector

__all__ = ["block"]

# TODO One thing one could still do to improve timings is implement a "fast einsum"
#      that does not call opt_einsum, but directly dispatches to libadcc. This could
#      lower the call overhead in the applies for the cases where we have only a
#      trivial einsum to do. For the moment I'm not convinced that is worth the
#      effort ... I suppose it only makes a difference for the cheaper ADC variants
#      (ADC(0), ADC(1), CVS-ADC(0-2)-x), but then on the other hand they are not
#      really so much our focus.


#
# Dispatch routine
#
"""
`apply` is a function mapping an AmplitudeVector to the contribution of this
block to the result of applying the ADC matrix. `diagonal` is an `AmplitudeVector`
containing the expression to the diagonal of the ADC matrix from this block.
"""
AdcBlock = namedtuple("AdcBlock", ["apply", "diagonal"])


def block(ground_state, spaces, order, variant=None, intermediates=None):
    """
    Gets ground state, potentially intermediates, spaces (ph, pphh and so on)
    and the perturbation theory order for the block,
    variant is "cvs" or sth like that.

    It is assumed largely, that CVS is equivalent to mp.has_core_occupied_space,
    while one would probably want in the long run that one can have an "o2" space,
    but not do CVS.
    """
    if isinstance(variant, str):
        variant = [variant]
    elif variant is None:
        variant = []
    reference_state = ground_state.reference_state
    if intermediates is None:
        intermediates = Intermediates(ground_state)

    if ground_state.has_core_occupied_space and "cvs" not in variant:
        raise ValueError("Cannot run a general (non-core-valence approximated) "
                         "ADC method on top of a ground state with a "
                         "core-valence separation.")
    if not ground_state.has_core_occupied_space and "cvs" in variant:
        raise ValueError("Cannot run a core-valence approximated ADC method on "
                         "top of a ground state without a "
                         "core-valence separation.")

    fn = "_".join(["block"] + variant + spaces + [str(order)])

    if fn not in globals():
        raise ValueError("Could not dispatch: "
                         f"spaces={spaces} order={order} variant={variant}. "
                         "Probably the secular matrix is not implemented for "
                         "the requested method.")
    return globals()[fn](reference_state, ground_state, intermediates)


#
# 0th order main
#
def block_ph_ph_0(hf, mp, intermediates):
    fCC = hf.fcc if hf.has_core_occupied_space else hf.foo
    diagonal = AmplitudeVector(ph=direct_sum("a-i->ia", hf.fvv.diagonal(),
                                             fCC.diagonal()))

    def apply(ampl):
        return AmplitudeVector(ph=(
            + einsum("ib,ab->ia", ampl.ph, hf.fvv)
            - einsum("IJ,Ja->Ia", fCC, ampl.ph)
        ))
    return AdcBlock(apply, diagonal)


block_cvs_ph_ph_0 = block_ph_ph_0


def diagonal_pphh_pphh_0(hf):
    # Note: adcman similarly does not symmetrise the occupied indices
    #       (for both CVS and general ADC)
    fCC = hf.fcc if hf.has_core_occupied_space else hf.foo
    res = direct_sum("-i-J+a+b->iJab",
                     hf.foo.diagonal(), fCC.diagonal(),
                     hf.fvv.diagonal(), hf.fvv.diagonal())
    return AmplitudeVector(pphh=res.symmetrise(2, 3))


def block_pphh_pphh_0(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(pphh=(
            + 2 * einsum("ijac,bc->ijab", ampl.pphh, hf.fvv).antisymmetrise(2, 3)
            - 2 * einsum("ik,kjab->ijab", hf.foo, ampl.pphh).antisymmetrise(0, 1)
        ))
    return AdcBlock(apply, diagonal_pphh_pphh_0(hf))


def block_cvs_pphh_pphh_0(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(pphh=(
            + 2 * einsum("iJac,bc->iJab", ampl.pphh, hf.fvv).antisymmetrise(2, 3)
            - einsum("ik,kJab->iJab", hf.foo, ampl.pphh)
            - einsum("JK,iKab->iJab", hf.fcc, ampl.pphh)
        ))
    return AdcBlock(apply, diagonal_pphh_pphh_0(hf))


#
# 0th order coupling
#
def block_ph_pphh_0(hf, mp, intermediates):
    return AdcBlock(lambda ampl: 0, 0)


def block_pphh_ph_0(hf, mp, intermediates):
    return AdcBlock(lambda ampl: 0, 0)


block_cvs_ph_pphh_0 = block_ph_pphh_0
block_cvs_pphh_ph_0 = block_pphh_ph_0


#
# 1st order main
#
def block_ph_ph_1(hf, mp, intermediates):
    fCC = hf.fcc if hf.has_core_occupied_space else hf.foo
    CvCv = hf.cvcv if hf.has_core_occupied_space else hf.ovov
    diagonal = AmplitudeVector(ph=(
        + direct_sum("a-i->ia", hf.fvv.diagonal(), fCC.diagonal())  # order 0
        - einsum("IaIa->Ia", CvCv)  # order 1
    ))

    def apply(ampl):
        return AmplitudeVector(ph=(                 # PT order
            + einsum("ib,ab->ia", ampl.ph, hf.fvv)  # 0
            - einsum("IJ,Ja->Ia", fCC, ampl.ph)     # 0
            - einsum("JaIb,Jb->Ia", CvCv, ampl.ph)  # 1
        ))
    return AdcBlock(apply, diagonal)


block_cvs_ph_ph_1 = block_ph_ph_1


def diagonal_pphh_pphh_1(hf):
    # Fock matrix and ovov diagonal term (sometimes called "intermediate diagonal")
    dinterm_ov = (direct_sum("a-i->ia", hf.fvv.diagonal(), hf.foo.diagonal())
                  - 2.0 * einsum("iaia->ia", hf.ovov)).evaluate()

    if hf.has_core_occupied_space:
        dinterm_Cv = (direct_sum("a-I->Ia", hf.fvv.diagonal(), hf.fcc.diagonal())
                      - 2.0 * einsum("IaIa->Ia", hf.cvcv)).evaluate()
        diag_oC = einsum("iJiJ->iJ", hf.ococ)
    else:
        dinterm_Cv = dinterm_ov
        diag_oC = einsum("ijij->ij", hf.oooo).symmetrise()

    diag_vv = einsum("abab->ab", hf.vvvv).symmetrise()
    return AmplitudeVector(pphh=(
        + direct_sum("ia+Jb->iJab", dinterm_ov, dinterm_Cv).symmetrise(2, 3)
        + direct_sum("iJ+ab->iJab", diag_oC, diag_vv)
    ))


def block_pphh_pphh_1(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(pphh=(  # 0th order
            + 2 * einsum("ijac,bc->ijab", ampl.pphh, hf.fvv).antisymmetrise(2, 3)
            - 2 * einsum("ik,kjab->ijab", hf.foo, ampl.pphh).antisymmetrise(0, 1)
            # 1st order
            + (
                -4 * einsum("ikac,kbjc->ijab", ampl.pphh, hf.ovov)
            ).antisymmetrise(0, 1).antisymmetrise(2, 3)
            + 0.5 * einsum("ijkl,klab->ijab", hf.oooo, ampl.pphh)
            + 0.5 * einsum("ijcd,abcd->ijab", ampl.pphh, hf.vvvv)
        ))
    return AdcBlock(apply, diagonal_pphh_pphh_1(hf))


def block_cvs_pphh_pphh_1(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(pphh=(
            # 0th order
            + 2.0 * einsum("iJac,bc->iJab", ampl.pphh, hf.fvv).antisymmetrise(2, 3)
            - 1.0 * einsum("ik,kJab->iJab", hf.foo, ampl.pphh)
            - 1.0 * einsum("JK,iKab->iJab", hf.fcc, ampl.pphh)
            # 1st order
            + (
                - 2.0 * einsum("iKac,KbJc->iJab", ampl.pphh, hf.cvcv)
                + 2.0 * einsum("icka,kJbc->iJab", hf.ovov, ampl.pphh)
            ).antisymmetrise(2, 3)
            + 1.0 * einsum("iJlK,lKab->iJab", hf.ococ, ampl.pphh)
            + 0.5 * einsum("iJcd,abcd->iJab", ampl.pphh, hf.vvvv)
        ))
    return AdcBlock(apply, diagonal_pphh_pphh_1(hf))


#
# 1st order coupling
#
def block_ph_pphh_1(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(ph=(
            + einsum("jkib,jkab->ia", hf.ooov, ampl.pphh)
            + einsum("ijbc,jabc->ia", ampl.pphh, hf.ovvv)
        ))
    return AdcBlock(apply, 0)


def block_cvs_ph_pphh_1(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(ph=(
            + sqrt(2) * einsum("jKIb,jKab->Ia", hf.occv, ampl.pphh)
            - 1 / sqrt(2) * einsum("jIbc,jabc->Ia", ampl.pphh, hf.ovvv)
        ))
    return AdcBlock(apply, 0)


def block_pphh_ph_1(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(pphh=(
            + einsum("ic,jcab->ijab", ampl.ph, hf.ovvv).antisymmetrise(0, 1)
            - einsum("ijka,kb->ijab", hf.ooov, ampl.ph).antisymmetrise(2, 3)
        ))
    return AdcBlock(apply, 0)


def block_cvs_pphh_ph_1(hf, mp, intermediates):
    def apply(ampl):
        return AmplitudeVector(pphh=(
            + sqrt(2) * einsum("jIKb,Ka->jIab",
                               hf.occv, ampl.ph).antisymmetrise(2, 3)
            - 1 / sqrt(2) * einsum("Ic,jcab->jIab", ampl.ph, hf.ovvv)
        ))
    return AdcBlock(apply, 0)


#
# 2nd order main
#
def block_ph_ph_2(hf, mp, intermediates):
    i1 = intermediates.adc2_i1
    i2 = intermediates.adc2_i2
    diagonal = AmplitudeVector(ph=(
        + direct_sum("a-i->ia", i1.diagonal(), i2.diagonal())
        - einsum("IaIa->Ia", hf.ovov)
        - einsum("ikac,ikac->ia", mp.t2oo, hf.oovv)
    ))

    # Not used anywhere else, so kept as an anonymous intermediate
    term_t2_eri = (
        + einsum("ijab,jkbc->ikac", mp.t2oo, hf.oovv)
        + einsum("ijab,jkbc->ikac", hf.oovv, mp.t2oo)
    ).evaluate()

    def apply(ampl):
        return AmplitudeVector(ph=(
            + einsum("ib,ab->ia", ampl.ph, i1)
            - einsum("ij,ja->ia", i2, ampl.ph)
            - einsum("jaib,jb->ia", hf.ovov, ampl.ph)    # 1
            - 0.5 * einsum("ikac,kc->ia", term_t2_eri, ampl.ph)  # 2
        ))
    return AdcBlock(apply, diagonal)


def block_cvs_ph_ph_2(hf, mp, intermediates):
    i1 = intermediates.adc2_i1
    diagonal = AmplitudeVector(ph=(
        + direct_sum("a-i->ia", i1.diagonal(), hf.fcc.diagonal())
        - einsum("IaIa->Ia", hf.cvcv)
    ))

    def apply(ampl):
        return AmplitudeVector(ph=(
            + einsum("ib,ab->ia", ampl.ph, i1)
            - einsum("ij,ja->ia", hf.fcc, ampl.ph)
            - einsum("JaIb,Jb->Ia", hf.cvcv, ampl.ph)
        ))
    return AdcBlock(apply, diagonal)


#
# 2nd order coupling
#
def block_ph_pphh_2(hf, mp, intermediates):
    pia_ooov = intermediates.adc3_pia
    pib_ovvv = intermediates.adc3_pib

    def apply(ampl):
        return AmplitudeVector(ph=(
            + einsum("jkib,jkab->ia", pia_ooov, ampl.pphh)
            + einsum("ijbc,jabc->ia", ampl.pphh, pib_ovvv)
            + einsum("icab,jkcd,jkbd->ia", hf.ovvv, ampl.pphh, mp.t2oo)  # 2nd
            + einsum("ijka,jlbc,klbc->ia", hf.ooov, mp.t2oo, ampl.pphh)  # 2nd
        ))
    return AdcBlock(apply, 0)


def block_cvs_ph_pphh_2(hf, mp, intermediates):
    pia_occv = intermediates.cvs_adc3_pia
    pib_ovvv = intermediates.adc3_pib

    def apply(ampl):
        return AmplitudeVector(ph=(1 / sqrt(2)) * (
            + 2.0 * einsum("lKIc,lKac->Ia", pia_occv, ampl.pphh)
            - einsum("lIcd,lacd->Ia", ampl.pphh, pib_ovvv)
            - einsum("jIKa,ljcd,lKcd->Ia", hf.occv, mp.t2oo, ampl.pphh)
        ))
    return AdcBlock(apply, 0)


def block_pphh_ph_2(hf, mp, intermediates):
    pia_ooov = intermediates.adc3_pia
    pib_ovvv = intermediates.adc3_pib

    def apply(ampl):
        return AmplitudeVector(pphh=(
            (
                + einsum("ic,jcab->ijab", ampl.ph, pib_ovvv)
                + einsum("lkic,kc,jlab->ijab", hf.ooov, ampl.ph, mp.t2oo)  # 2st
            ).antisymmetrise(0, 1)
            + (
                - einsum("ijka,kb->ijab", pia_ooov, ampl.ph)
                - einsum("ijac,kbcd,kd->ijab", mp.t2oo, hf.ovvv, ampl.ph)  # 2st
            ).antisymmetrise(2, 3)
        ))
    return AdcBlock(apply, 0)


def block_cvs_pphh_ph_2(hf, mp, intermediates):
    pia_occv = intermediates.cvs_adc3_pia
    pib_ovvv = intermediates.adc3_pib

    def apply(ampl):
        return AmplitudeVector(pphh=(1 / sqrt(2)) * (
            - 2.0 * einsum("jIKa,Kb->jIab", pia_occv, ampl.ph).antisymmetrise(2, 3)
            - einsum("Ic,jcab->jIab", ampl.ph, pib_ovvv)
            - einsum("lKIc,Kc,jlab->jIab", hf.occv, ampl.ph, mp.t2oo)
        ))
    return AdcBlock(apply, 0)


#
# 3rd order main
#
def block_ph_ph_3(hf, mp, intermediates):
    if hf.has_core_occupied_space:
        m11 = intermediates.cvs_adc3_m11
    else:
        m11 = intermediates.adc3_m11
    diagonal = AmplitudeVector(ph=einsum("iaia->ia", m11))

    def apply(ampl):
        return AmplitudeVector(ph=einsum("iajb,jb->ia", m11, ampl.ph))
    return AdcBlock(apply, diagonal)


block_cvs_ph_ph_3 = block_ph_ph_3


#
# Intermediates
#

@register_as_intermediate
def adc2_i1(hf, mp, intermediates):
    # This definition differs from libadc. It additionally has the hf.fvv term.
    return hf.fvv + 0.5 * einsum("ijac,ijbc->ab", mp.t2oo, hf.oovv).symmetrise()


@register_as_intermediate
def adc2_i2(hf, mp, intermediates):
    # This definition differs from libadc. It additionally has the hf.foo term.
    return hf.foo - 0.5 * einsum("ikab,jkab->ij", mp.t2oo, hf.oovv).symmetrise()


def adc3_i1(hf, mp, intermediates):
    # Used for both CVS and general
    td2 = mp.td2(b.oovv)
    p0 = mp.second_order_dm_correction(apply_cvs=hf.has_core_occupied_space)

    t2eri_sum = (
        + einsum("jicb->ijcb", mp.t2eri(b.oovv, b.ov))  # t2eri4
        - 0.25 * mp.t2eri(b.oovv, b.vv)                 # t2eri5
    )
    return (
        (  # symmetrise a<>b
            + 0.5 * einsum("ijac,ijbc->ab", mp.t2oo + td2, hf.oovv)
            - 1.0 * einsum("ijac,ijcb->ab", mp.t2oo, t2eri_sum)
            - 2.0 * einsum("iabc,ic->ab", hf.ovvv, p0.ov)
        ).symmetrise()
        + einsum("iajb,ij->ab", hf.ovov, p0.oo)
        + einsum("acbd,cd->ab", hf.vvvv, p0.vv)
    )


def adc3_i2(hf, mp, intermediates):
    # Used only for general
    td2 = mp.td2(b.oovv)
    p0 = mp.mp2_diffdm

    # t2eri4 + t2eri3 / 4
    t2eri_sum = mp.t2eri(b.oovv, b.ov) + 0.25 * mp.t2eri(b.oovv, b.oo)
    return (
        (  # symmetrise i<>j
            + 0.5 * einsum("ikab,jkab->ij", mp.t2oo + td2, hf.oovv)
            - 1.0 * einsum("ikab,jkab->ij", mp.t2oo, t2eri_sum)
            + 2.0 * einsum("kija,ka->ij", hf.ooov, p0.ov)
        ).symmetrise()
        - einsum("ikjl,kl->ij", hf.oooo, p0.oo)
        - einsum("iajb,ab->ij", hf.ovov, p0.vv)
    )


def cvs_adc3_i2(hf, mp, intermediates):
    cvs_p0 = mp.second_order_dm_correction(apply_cvs=True)
    return (
        + 2.0 * einsum("kIJa,ka->IJ", hf.occv, cvs_p0.ov).symmetrise()
        - 1.0 * einsum("kIlJ,kl->IJ", hf.ococ, cvs_p0.oo)
        - 1.0 * einsum("IaJb,ab->IJ", hf.cvcv, cvs_p0.vv)
    )


@register_as_intermediate
def adc3_m11(hf, mp, intermediates):
    td2 = mp.td2(b.oovv)
    p0 = mp.mp2_diffdm

    i1 = adc3_i1(hf, mp, intermediates).evaluate()
    i2 = adc3_i2(hf, mp, intermediates).evaluate()
    t2sq = einsum("ikac,jkbc->iajb", mp.t2oo, mp.t2oo).evaluate()

    # Build two Kronecker deltas
    d_oo = zeros_like(hf.foo)
    d_vv = zeros_like(hf.fvv)
    d_oo.set_mask("ii", 1.0)
    d_vv.set_mask("aa", 1.0)

    t2eri_sum = (
        + 2.0 * mp.t2eri(b.oovv, b.ov).symmetrise((0, 1), (2, 3))  # t2eri4
        + 0.5 * mp.t2eri(b.oovv, b.vv)                             # t2eri5
        + 0.5 * mp.t2eri(b.oovv, b.oo)                             # t2eri3
    )
    return (
        + einsum("ij,ab->iajb", d_oo, hf.fvv + i1)
        - einsum("ij,ab->iajb", hf.foo - i2, d_vv)
        - einsum("jaib->iajb", hf.ovov)
        - (  # symmetrise i<>j and a<>b
            + einsum("jkbc,ikac->iajb", hf.oovv, mp.t2oo + td2)
            - einsum("jkbc,ikac->iajb", mp.t2oo, t2eri_sum)
            - einsum("ibac,jc->iajb", hf.ovvv, 2.0 * p0.ov)
            - einsum("ikja,kb->iajb", hf.ooov, 2.0 * p0.ov)
            - einsum("jaic,bc->iajb", hf.ovov, p0.vv)
            + einsum("ik,jakb->iajb", p0.oo, hf.ovov)
            + einsum("ibkc,kajc->iajb", hf.ovov, 2.0 * t2sq)
        ).symmetrise((0, 2), (1, 3))
        # TODO This hack is done to avoid opt_einsum being smart and instantiating
        #      a tensor of dimension 6 (to avoid the vvvv tensor) in some cases,
        #      which is the right thing to do, but not yet supported.
        # + 0.5 * einsum("icjd,klac,klbd->iajb", hf.ovov, mp.t2oo, mp.t2oo)
        + 0.5 * einsum("icjd,acbd->iajb", hf.ovov,
                       einsum("klac,klbd->acbd", mp.t2oo, mp.t2oo))
        # + 0.5 * einsum("ikcd,jlcd,kalb->iajb", mp.t2oo, mp.t2oo, hf.ovov)
        + 0.5 * einsum("ikjl,kalb->iajb",
                       einsum("ikcd,jlcd->ikjl", mp.t2oo, mp.t2oo), hf.ovov)
        - einsum("iljk,kalb->iajb", hf.oooo, t2sq)
        - einsum("idjc,acbd->iajb", t2sq, hf.vvvv)
    )


@register_as_intermediate
def cvs_adc3_m11(hf, mp, intermediates):
    i1 = adc3_i1(hf, mp, intermediates).evaluate()
    i2 = cvs_adc3_i2(hf, mp, intermediates).evaluate()
    t2sq = einsum("ikac,jkbc->iajb", mp.t2oo, mp.t2oo).evaluate()
    cvs_p0 = mp.second_order_dm_correction(apply_cvs=True)

    # Build two Kronecker deltas
    d_cc = zeros_like(hf.fcc)
    d_vv = zeros_like(hf.fvv)
    d_cc.set_mask("II", 1.0)
    d_vv.set_mask("aa", 1.0)

    return (
        + einsum("IJ,ab->IaJb", d_cc, hf.fvv + i1)
        - einsum("IJ,ab->IaJb", hf.fcc - i2, d_vv)
        - einsum("JaIb->IaJb", hf.cvcv)
        + (  # symmetrise I<>J and a<>b
            + einsum("JaIc,bc->IaJb", hf.cvcv, cvs_p0.vv)
            - einsum("kIJa,kb->IaJb", hf.occv, 2.0 * cvs_p0.ov)
        ).symmetrise((0, 2), (1, 3))
        # TODO This hack is done to avoid opt_einsum being smart and instantiating
        #      a tensor of dimension 6 (to avoid the vvvv tensor) in some cases,
        #      which is the right thing to do, but not yet supported.
        # + 0.5 * einsum("IcJd,klac,klbd->IaJb", hf.cvcv, mp.t2oo, mp.t2oo)
        + 0.5 * einsum("IcJd,acbd->IaJb", hf.cvcv,
                       einsum("klac,klbd->acbd", mp.t2oo, mp.t2oo))
        - einsum("lIkJ,kalb->IaJb", hf.ococ, t2sq)
    )


@register_as_intermediate
def adc3_pia(hf, mp, intermediates):
    # This definition differs from libadc. It additionally has the hf.ooov term.
    return (                          # Perturbation theory in ADC coupling block
        + hf.ooov                                            # 1st order
        - 2.0 * mp.t2eri(b.ooov, b.ov).antisymmetrise(0, 1)  # 2nd order
        - 0.5 * mp.t2eri(b.ooov, b.vv)                       # 2nd order
    )


@register_as_intermediate
def cvs_adc3_pia(hf, mp, intermediates):
    # Perturbation theory in CVS-ADC coupling block:
    #       1st                     2nd
    return hf.occv - einsum("jlac,lKIc->jIKa", mp.t2oo, hf.occv)


@register_as_intermediate
def adc3_pib(hf, mp, intermediates):
    # This definition differs from libadc. It additionally has the hf.ovvv term.
    return (                          # Perturbation theory in ADC coupling block
        + hf.ovvv                                            # 1st order
        + 2.0 * mp.t2eri(b.ovvv, b.ov).antisymmetrise(2, 3)  # 2nd order
        - 0.5 * mp.t2eri(b.ovvv, b.oo)                       # 2nd order
    )



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
from pyscf import gto, scf

import adcc

# Run SCF in pyscf
mol = gto.M(
    atom='S  -0.38539679062   0 -0.27282082253;'
         'H  -0.0074283962687 0  2.2149138578;'
         'H   2.0860198029    0 -0.74589639249',
    basis='cc-pvtz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-13
scfres.kernel()

print(adcc.banner())

# Run an adc2x calculation:
singlets = adcc.cvs_adc2x(scfres, core_orbitals=1, frozen_core=1, n_singlets=3)
triplets = adcc.cvs_adc2x(singlets.matrix, n_triplets=3)

print(singlets.describe())
print()
print(triplets.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2026 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .NParticleOperator import NParticleOperator, OperatorSymmetry


class NParticleDensity(NParticleOperator):
    def __init__(self, spaces, n_particle_op, symmetry=OperatorSymmetry.HERMITIAN):
        """
        General N-particle density operator.

        Density matrices require a different MO-to-AO transformation than general
        operators due to the non-orthonormal AO basis (C^\\dagger S C= 1).
        While general operators transform as

        d_MO = C^\\dagger S d_AO S C

        density matrices transform as

        D_MO = C^\\dagger d_AO C

        Expectation values (inner products) of a density and an operator are
        identical in the AO and MO basis. Note that it is important that only one
        of them is transformed using the overlap matrix!

        It therefore makes sense to consistently choose the transformations of the
        Fock matrix and the HF density matrix for all operators and densities,
        respectively.
        """
        super().__init__(spaces, n_particle_op, symmetry=symmetry)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
"""Example computing the TPA cross section for water using ADC
(10.1063/1.3682324)
"""
import adcc
import numpy as np

from pyscf import gto, scf
from adcc.solver.preconditioner import JacobiPreconditioner
from adcc.solver import IndexSymmetrisation
from adcc.solver.conjugate_gradient import conjugate_gradient, default_print
from adcc.adc_pp.modified_transition_moments import modified_transition_moments
from adcc.IsrMatrix import IsrMatrix


class ShiftedMat(adcc.AdcMatrix):
    def __init__(self, method, mp_results, omega=0.0):
        self.omega = omega
        super().__init__(method, mp_results)
        self.omegamat = adcc.ones_like(self.diagonal()) * omega

    def __matmul__(self, other):
        return super().__matmul__(other) - self.omegamat * other


# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='aug-cc-pvdz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-12
scfres.conv_tol_grad = 1e-9
scfres.kernel()

# solve for Eigenvalues
state = adcc.adc2(scfres, n_singlets=1, conv_tol=1e-8)

# setup modified transition moments
dips = state.reference_state.operators.electric_dipole
rhss = modified_transition_moments("adc2", state.ground_state, dips)
isrmatrix = IsrMatrix("adc2", state.ground_state, dips)

S = np.zeros((len(state.excitation_energy), 3, 3))
for f, ee in enumerate(state.excitation_energy):
    freq = ee / 2.0
    matrix = ShiftedMat("adc2", state.ground_state, freq)
    preconditioner = JacobiPreconditioner(matrix)
    explicit_symmetrisation = IndexSymmetrisation(matrix)
    preconditioner.update_shifts(freq)
    response = []
    # solve all systems of equations
    for mu in range(3):
        rhs = rhss[mu]
        x0 = preconditioner.apply(rhs)
        res = conjugate_gradient(
            matrix, rhs=rhs, x0=x0, callback=default_print,
            Pinv=preconditioner, conv_tol=1e-6,
            explicit_symmetrisation=explicit_symmetrisation
        )
        response.append(res)
    right_vec = isrmatrix @ state.excitation_vector[f]
    for mu in range(3):
        for nu in range(mu, 3):
            # compute the matrix element
            S[f, mu, nu] = (
                response[mu].solution @ right_vec[nu]
                + response[nu].solution @ right_vec[mu]
            )
            S[f, nu, mu] = S[f, mu, nu]
    print("Two-Photon Matrix for state", f)
    print(S[f])
    delta = 1.0 / 15.0 * (
        np.einsum('mm,vv->', S[f], S[f])
        + np.einsum('mv,mv->', S[f], S[f])
        + np.einsum('mv,vm->', S[f], S[f])
    )
    print("TPA Cross section [a.u.]: {:.4f}".format(delta))
    np.testing.assert_allclose(6.5539, delta, atol=1e-4)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2026 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .. import block as b
from ..LazyMp import LazyMp
from ..AdcMethod import IsrMethod
from ..functions import einsum, zeros_like
from ..Intermediates import Intermediates
from ..AmplitudeVector import AmplitudeVector
from ..TwoParticleDensity import TwoParticleDensity
from ..NParticleOperator import OperatorSymmetry
from .util import (
    check_doubles_amplitudes, check_singles_amplitudes, check_triples_amplitudes
)

from typing import Optional, Union


def diffdm_isr0_2p(mp: LazyMp, amplitude: AmplitudeVector,
                   intermediates: Intermediates) -> TwoParticleDensity:
    check_singles_amplitudes([b.o, b.v], amplitude)
    u1 = amplitude.ph

    hf = mp.reference_state
    d_oo = zeros_like(hf.foo)
    d_oo.set_mask("ii", 1)

    dm = TwoParticleDensity(mp, symmetry=OperatorSymmetry.HERMITIAN)

    # one-particle ISR(0) diffdm
    p1_oo = -einsum("ia,la->il", u1, u1).evaluate()
    p1_vv = einsum("ka,kb->ab", u1, u1).evaluate()

    dm.oooo = (
        # N^4: O^4 / N^4: O^4
        - 4.0 * einsum("il,jk->ijkl", p1_oo, d_oo)
    ).antisymmetrise(0, 1).antisymmetrise(2, 3)
    dm.ovov = (
        # N^4: O^2V^2 / N^4: O^2V^2
        - 1.0 * einsum("ja,ib->iajb", u1, u1)
        # N^4: O^2V^2 / N^4: O^2V^2
        + 1.0 * einsum("ab,ij->iajb", p1_vv, d_oo)
    )
    return dm


def diffdm_isr1s_2p(mp: LazyMp, amplitude: AmplitudeVector,
                    intermediates: Intermediates) -> TwoParticleDensity:
    dm = diffdm_isr0_2p(mp, amplitude, intermediates)  # Get ISR(0) result
    u1 = amplitude.ph

    hf = mp.reference_state
    d_oo = zeros_like(hf.foo)
    d_oo.set_mask("ii", 1)

    t2 = mp.t2(b.oovv)
    # TODO move to intermediates!
    # one-particle ISR(0) diffdm
    p1_oo = -einsum("ia,la->il", u1, u1).evaluate()

    # ISR(2) ISR intermediate (TODO Move to intermediates)
    ru1 = einsum("ijab,jb->ia", t2, u1).evaluate()
    # new ones
    ru1_ooov = einsum("kc,ijac->ijka", u1, t2).evaluate()

    dm.oovv += (
        # N^4: O^2V^2 / N^4: O^2V^2
        + 4.0 * (
            einsum("ib,ja->ijab", ru1, u1)
        ).antisymmetrise(0, 1).antisymmetrise(2, 3)
        # N^5: O^3V^2 / N^4: O^2V^2
        + 2.0 * (
            einsum("ijka,kb->ijab", ru1_ooov, u1)
        ).antisymmetrise(2, 3)
        # N^5: O^3V^2 / N^4: O^2V^2
        - 2.0 * (
            einsum("jk,ikab->ijab", p1_oo, t2)
        ).antisymmetrise(0, 1)
    )

    return dm


def diffdm_isr1_2p(mp: LazyMp, amplitude: AmplitudeVector,
                   intermediates: Intermediates) -> TwoParticleDensity:
    dm = diffdm_isr1s_2p(mp, amplitude, intermediates)  # Get ISR(1)-s result

    try:
        # ISR(1)-d
        check_doubles_amplitudes([b.o, b.o, b.v, b.v], amplitude)
        u1, u2 = amplitude.ph, amplitude.pphh

        hf = mp.reference_state
        d_oo = zeros_like(hf.foo)
        d_oo.set_mask("ii", 1)

        dm.ooov += (
            # N^5: O^3V^2 / N^4: O^2V^2
            - 2.0 * einsum("kb,ijab->ijka", u1, u2)
            # N^4: O^2V^2 / N^4: O^2V^2
            - 4.0 * einsum(
                "ja,ik->ijka", einsum("lb,jlab->ja", u1, u2), d_oo
            ).antisymmetrise(0, 1)
        )

        dm.ovvv += (
            # N^5: O^2V^3 / N^4: O^1V^3
            - 2.0 * einsum("ja,ijbc->iabc", u1, u2)
        )
    except ValueError:
        # no doubles contribution
        pass
    return dm


def diffdm_isr2d_2p(mp: LazyMp, amplitude: AmplitudeVector,
                    intermediates: Intermediates) -> TwoParticleDensity:
    dm = diffdm_isr1_2p(mp, amplitude, intermediates)  # Get ISR(1) result
    check_doubles_amplitudes([b.o, b.o, b.v, b.v], amplitude)
    u1, u2 = amplitude.ph, amplitude.pphh
    hf = mp.reference_state
    d_oo = zeros_like(hf.foo)
    d_oo.set_mask("ii", 1)

    t2 = mp.t2(b.oovv)
    td2 = mp.td2(b.oovv)
    p0 = mp.mp2_diffdm

    dm.oooo += (
        # N^6: O^4V^2 / N^4: O^2V^2
        + 2.0 * einsum("ijab,klab->ijkl", u2, u2)
        # N^6: O^5V^1 / N^4: O^2V^2
        - 1.0 * einsum("klmc,ijmc->ijkl", einsum("mb,klbc->klmc", u1, t2),
                       einsum("ma,ijac->ijmc", u1, t2))
        + (
            # N^5: O^3V^2 / N^4: O^2V^2
            - 8.0 * einsum("ik,jl->ijkl", einsum("imab,kmab->ik", u2, u2), d_oo)
            # N^4: O^4 / N^4: O^4
            + 4.0 * einsum("il,jk->ijkl", einsum("ia,la->il", u1, u1), p0.oo)
            # N^6: O^5V^1 / N^4: O^2V^2
            + 4.0 * einsum("jlmc,ikmc->ijkl", einsum("lb,jmbc->jlmc", u1, t2),
                           einsum("ia,kmac->ikmc", u1, t2))
            # N^4: O^2V^2 / N^4: O^2V^2
            + 4.0 * einsum("il,jk->ijkl",
                           einsum("lc,ic->il", einsum("nb,lnbc->lc", u1, t2),
                                  einsum("ma,imac->ic", u1, t2)), d_oo)
            # N^5: O^3V^2 / N^4: O^2V^2
            + 4.0 * einsum("ik,jl->ijkl", einsum("kmnc,imnc->ik",
                                                 einsum("nb,kmbc->kmnc", u1, t2),
                                                 einsum("na,imac->imnc", u1, t2)),
                           d_oo)
            # N^5: O^3V^2 / N^4: O^2V^2
            + 2.0 * einsum("ik,jl->ijkl",
                           einsum("inbc,knbc->ik",
                                  einsum("mn,imbc->inbc",
                                         einsum("ia,la->il", u1, u1), t2), t2),
                           d_oo)
        ).antisymmetrise(0, 1).antisymmetrise(2, 3)
        + (
            # N^5: O^3V^2 / N^4: O^2V^2
            + 4.0 * einsum("iklb,jb->ijkl",
                           einsum("ic,klbc->iklb", einsum("ma,imac->ic", u1, t2),
                                  t2), u1)
            # N^6: O^4V^2 / N^4: O^2V^2
            + 2.0 * einsum("jklm,im->ijkl",
                           einsum("jmbc,klbc->jklm", t2, t2),
                           einsum("ia,la->il", u1, u1))
        ).antisymmetrise(0, 1).symmetrise([(0, 2), (1, 3)])
        + (
            # N^4: O^4 / N^4: O^4
            + 4.0 * einsum("il,jk->ijkl", einsum("im,lm->il",
                                                 einsum("ia,la->il", u1, u1),
                                                 p0.oo), d_oo)
            # N^4: O^2V^2 / N^4: O^2V^2
            + 4.0 * einsum("ik,jl->ijkl",
                           einsum("kb,ib->ik",
                                  einsum("mc,kmbc->kb",
                                         einsum("na,mnac->mc", u1, t2), t2), u1),
                           d_oo)
        ).antisymmetrise(0, 1).antisymmetrise(2, 3).symmetrise([(0, 2), (1, 3)])
    )
    dm.ooov += (
        # N^6: O^4V^2 / N^4: O^2V^2
        + 1.0 * einsum("ijkl,la->ijka", einsum("klbc,ijbc->ijkl", u2, t2), u1)
        # N^5: O^3V^2 / N^4: O^2V^2
        + 2.0 * einsum("kb,ijab->ijka", einsum("lc,klbc->kb", u1, u2), t2)
        + (
            # N^5: O^3V^2 / N^4: O^2V^2
            + 2.0 * einsum("jk,ia->ijka", einsum("klbc,jlbc->jk", u2, t2), u1)
            # N^6: O^4V^2 / N^4: O^2V^2
            - 4.0 * einsum("jklb,ilab->ijka", einsum("jc,klbc->jklb", u1, u2), t2)
            # N^5: O^3V^2 / N^4: O^2V^2
            + 2.0 * einsum("ja,ik->ijka",
                           einsum("jlmb,lmab->ja",
                                  einsum("jc,lmbc->jlmb", u1, u2), t2), d_oo)
            # N^4: O^2V^2 / N^4: O^2V^2
            - 4.0 * einsum("ia,jk->ijka",
                           einsum("lb,ilab->ia",
                                  einsum("mc,lmbc->lb", u1, u2), t2), d_oo)
            # N^5: O^3V^2 / N^4: O^2V^2
            + 2.0 * einsum("ia,jk->ijka",
                           einsum("il,la->ia",
                                  einsum("lmbc,imbc->il", u2, t2), u1), d_oo)
            # N^4: O^3V^1 / N^4: O^3V^1
            + 2.0 * einsum("jk,ia->ijka", einsum("kb,jb->jk", u1, p0.ov), u1)
            # N^4: O^3V^1 / N^4: O^3V^1
            + 2.0 * einsum("jk,ia->ijka", einsum("jb,kb->jk", u1, u1), p0.ov)
            # N^4: O^3V^1 / N^4: O^3V^1
            + 2.0 * einsum("ia,jk->ijka",
                           einsum("il,la->ia",
                                  einsum("lb,ib->il", u1, p0.ov), u1), d_oo)
            # N^4: O^3V^1 / N^4: O^3V^1
            + 2.0 * einsum("ia,jk->ijka",
                           einsum("il,la->ia", einsum("ia,la->il", u1, u1),
                                  p0.ov), d_oo)
        ).antisymmetrise(0, 1)
    )
    dm.oovv += (
        # N^4: O^2V^2 / N^4: O^2V^2
        + 4.0 * einsum("ib,ja->ijab",
                       einsum("kc,ikbc->ib", u1, td2),
                       u1).antisymmetrise(0, 1).antisymmetrise(2, 3)
        # N^5: O^3V^2 / N^4: O^2V^2
        + 2.0 * einsum("ijka,kb->ijab",
                       einsum("kc,ijac->ijka", u1, td2), u1).antisymmetrise(2, 3)
        # N^5: O^3V^2 / N^4: O^2V^2
        + 2.0 * einsum("jk,ikab->ijab", einsum("jc,kc->jk", u1, u1), td2)
    )
    dm.ovov += (
        # N^6: O^3V^3 / N^4: O^2V^2
        - 4.0 * einsum("jkac,ikbc->iajb", u2, u2)
        # N^5: O^2V^3 / N^4: O^2V^2
        + 2.0 * einsum("ab,ij->iajb", einsum("klac,klbc->ab", u2, u2), d_oo)
        # N^6: O^4V^2 / N^4: O^2V^2
        + 1.0 * einsum("iklb,jkla->iajb", einsum("ld,ikbd->iklb", u1, t2),
                       einsum("lc,jkac->jkla", u1, t2))
        # N^4: O^2V^2 / N^4: O^2V^2
        + 1.0 * einsum("ab,ij->iajb", einsum("ka,kb->ab", u1, u1), p0.oo)
        # N^4: O^2V^2 / N^4: O^2V^2
        - 1.0 * einsum("ij,ab->iajb", einsum("ic,jc->ij", u1, u1), p0.vv)
        # N^6: O^3V^3 / N^4: O^2V^2
        + 1.0 * einsum("jlac,ilbc->iajb",
                       einsum("kl,jkac->jlac", einsum("kd,ld->kl", u1, u1), t2), t2)
        # N^6: O^4V^2 / N^4: O^2V^2
        + 0.5 * einsum("ijla,lb->iajb",
                       einsum("ijkl,ka->ijla",
                              einsum("ikcd,jlcd->ijkl", t2, t2), u1), u1)
        # N^6: O^4V^2 / N^4: O^2V^2
        + 0.5 * einsum("jklb,ikla->iajb",
                       einsum("jd,klbd->jklb", u1, t2),
                       einsum("ic,klac->ikla", u1, t2))
        # N^4: O^2V^2 / N^4: O^2V^2
        - 1.0 * einsum("ib,ja->iajb",
                       einsum("kd,ikbd->ib", u1, t2), einsum("lc,jlac->ja", u1, t2))
        # N^5: O^2V^3 / N^4: O^2V^2
        + 1.0 * einsum("ab,ij->iajb",
                       einsum("lmad,lmbd->ab",
                              einsum("km,klad->lmad",
                                     einsum("kc,mc->km", u1, u1), t2), t2), d_oo)
        # N^4: O^2V^2 / N^4: O^2V^2
        - 1.0 * einsum("ab,ij->iajb",
                       einsum("lb,la->ab",
                              einsum("kd,klbd->lb", u1, t2),
                              einsum("mc,lmac->la", u1, t2)), d_oo)
        # N^5: O^3V^2 / N^4: O^2V^2
        - 0.5 * einsum("ab,ij->iajb",
                       einsum("klmb,klma->ab",
                              einsum("md,klbd->klmb", u1, t2),
                              einsum("mc,klac->klma", u1, t2)), d_oo)
        + (
            # N^4: O^2V^2 / N^4: O^2V^2
            + 1.0 * einsum("ib,ja->iajb", einsum("ic,bc->ib", u1, p0.vv), u1)
            # N^4: O^2V^2 / N^4: O^2V^2
            - 1.0 * einsum("ib,ja->iajb", einsum("kb,ik->ib", u1, p0.oo), u1)
            # N^5: O^3V^2 / N^4: O^2V^2
            + 2.0 * einsum("ijlb,la->iajb",
                           einsum("jc,ilbc->ijlb",
                                  einsum("kd,jkcd->jc", u1, t2), t2), u1)
            # N^5: O^3V^2 / N^4: O^2V^2
            + 2.0 * einsum("kb,ijka->iajb",
                           einsum("ld,klbd->kb", u1, t2),
                           einsum("ic,jkac->ijka", u1, t2))
            # N^6: O^4V^2 / N^4: O^2V^2
            - 2.0 * einsum("ijkb,ka->iajb",
                           einsum("jklc,ilbc->ijkb",
                                  einsum("kd,jlcd->jklc", u1, t2), t2), u1)
            # N^6: O^4V^2 / N^4: O^2V^2
            - 2.0 * einsum("ijlb,la->iajb",
                           einsum("ijkc,klbc->ijlb",
                                  einsum("id,jkcd->ijkc", u1, t2), t2), u1)
            # N^6: O^3V^3 / N^4: O^2V^2
            - 2.0 * einsum("ikbc,jkac->iajb",
                           einsum("il,klbc->ikbc",
                                  einsum("id,ld->il", u1, u1), t2), t2)
            # N^4: O^2V^2 / N^4: O^2V^2
            - 1.0 * einsum("ib,ja->iajb",
                           einsum("kc,ikbc->ib",
                                  einsum("ld,klcd->kc", u1, t2), t2), u1)
            # N^4: O^2V^2 / N^4: O^2V^2
            - 1.0 * einsum("ab,ij->iajb",
                           einsum("kb,ka->ab",
                                  einsum("kc,bc->kb", u1, p0.vv), u1), d_oo)
            # N^4: O^2V^2 / N^4: O^2V^2
            - 1.0 * einsum("ab,ij->iajb",
                           einsum("mb,ma->ab",
                                  einsum("ld,lmbd->mb",
                                         einsum("kc,klcd->ld", u1, t2),
                                         t2), u1), d_oo)
        ).symmetrise([(0, 2), (1, 3)])
    )
    dm.ovvv += (
        # N^6: O^3V^3 / N^4: O^1V^3
        + 1.0 * einsum("ijka,jkbc->iabc", einsum("id,jkad->ijka", u1, u2), t2)
        # N^5: O^2V^3 / N^4: O^1V^3
        + 2.0 * einsum("ja,ijbc->iabc", einsum("kd,jkad->ja", u1, u2), t2)
        + (
            # N^5: O^2V^3 / N^4: O^1V^3
            + 2.0 * einsum("ac,ib->iabc", einsum("jkad,jkcd->ac", u2, t2), u1)
            # N^6: O^3V^3 / N^4: O^1V^3
            - 4.0 * einsum("ikab,kc->iabc", einsum("jkad,ijbd->ikab", u2, t2), u1)
            # N^4: O^1V^3 / N^4: O^1V^3
            + 2.0 * einsum("ac,ib->iabc", einsum("ja,jc->ac", u1, p0.ov), u1)
            # N^4: O^1V^3 / N^4: O^1V^3
            + 2.0 * einsum("ac,ib->iabc", einsum("ja,jc->ac", u1, u1), p0.ov)
        ).antisymmetrise(2, 3)
    )
    dm.vvvv += (
        # N^6: O^2V^4 / N^4: V^4
        + 2 * einsum("ijab,ijcd->abcd", u2, u2)
        # N^6: O^2V^4 / N^4: V^4
        - 1 * einsum("jkab,jkcd->abcd",
                     einsum("ij,ikab->jkab", einsum("ie,je->ij", u1, u1), t2), t2)
        + (
            # N^4: V^4 / N^4: V^4
            + 4.0 * einsum("ac,bd->abcd", einsum("ia,ic->ac", u1, u1), p0.vv)
            # N^6: O^3V^3 / N^4: V^4
            + 4.0 * einsum("jabc,jd->abcd",
                           einsum("ijbc,ia->jabc",
                                  einsum("jkbe,ikce->ijbc", t2, t2), u1), u1)
        ).antisymmetrise(0, 1).antisymmetrise(2, 3)
        + (
            # N^5: O^1V^4 / N^4: V^4
            + 4.0 * einsum("ka,kbcd->abcd",
                           einsum("ie,ikae->ka", u1, t2),
                           einsum("jb,jkcd->kbcd", u1, t2))
            # N^6: O^3V^3 / N^4: V^4
            + 2.0 * einsum("ibcd,ia->abcd",
                           einsum("ijkb,jkcd->ibcd",
                                  einsum("ie,jkbe->ijkb", u1, t2), t2), u1)
        ).antisymmetrise(0, 1).symmetrise([(0, 2), (1, 3)])
    )
    return dm


def diffdm_isr2_2p(mp: LazyMp, amplitude: AmplitudeVector,
                   intermediates: Intermediates) -> TwoParticleDensity:
    dm = diffdm_isr2d_2p(mp, amplitude, intermediates)  # Get ISR(2)-d result
    # evaluate additional contributions from the S-T block
    # if the vector has a triples component
    try:
        check_triples_amplitudes([b.o, b.o, b.o, b.v, b.v, b.v], amplitude)
    except ValueError:
        return dm
    u1, u3 = amplitude.ph, amplitude.ppphhh
    # N^6: O^3V^3 / N^6: O^3V^3
    dm.oovv += - 6 * einsum("kc,ijkabc->ijab", u1, u3)
    return dm


# dict controlling the dispatch of the state_diffdm function
DISPATCH = {
    "isr0": diffdm_isr0_2p,
    "isr1s": diffdm_isr1s_2p,
    "isr1": diffdm_isr1_2p,
    "isr2d": diffdm_isr2d_2p,
    "isr2": diffdm_isr2_2p,
}


def state_diffdm_2p(method: Union[str, IsrMethod], ground_state: LazyMp,
                    amplitude: AmplitudeVector,
                    intermediates: Optional[Intermediates] = None):
    """
    Compute the two-particle difference density matrix of an excited state
    in the MO basis.

    Parameters
    ----------
    method : str, IsrMethod
        The method to use for the computation (e.g. "isr2")
    ground_state : LazyMp
        The ground state upon which the excitation was based
    amplitude : AmplitudeVector
        The amplitude vector
    intermediates : adcc.Intermediates
        Intermediates from the ADC calculation to reuse
    """
    if not isinstance(method, IsrMethod):
        method = IsrMethod(method)
    if not isinstance(ground_state, LazyMp):
        raise TypeError("ground_state should be a LazyMp object.")
    if not isinstance(amplitude, AmplitudeVector):
        raise TypeError("amplitude should be an AmplitudeVector object.")
    if intermediates is None:
        intermediates = Intermediates(ground_state)

    if method.name not in DISPATCH:
        raise NotImplementedError("state_diffdm_2p is not implemented "
                                  f"for {method.name}.")
    else:
        ret = DISPATCH[method.name](ground_state, amplitude, intermediates)
        return ret.evaluate()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import warnings
import numpy as np
import scipy.linalg as la

from adcc import evaluate, lincomb
from adcc.timings import Timer
from adcc.AmplitudeVector import AmplitudeVector

from .orthogonaliser import GramSchmidtOrthogonaliser


class LanczosIterator:
    def __init__(self, matrix, guesses, ritz_vectors=None, ritz_values=None,
                 ritz_overlaps=None, explicit_symmetrisation=None):
        """
        Initialise an iterator generating :py:class:`LanczosSubspace` objects,
        which represent a growing Krylov subspace started from the `matrix`
        and guess vectors `guesses` or optionally using a thick restart from
        the passed Ritz vectors, Ritz values and Ritz overlaps.

        Parameters
        ----------
        matrix
            Matrix to build the Krylov subspace
        guesses : list
            Vectors to build the Krylov subspace
        ritz_vectors : list or NoneType, optional
            Ritz vectors for thick restarts
        ritz_values : numpy.ndarray or NoneType, optional
            Ritz values corresponding to the `ritz_vectors` for thick restarts
        ritz_overlaps : numpy.ndarray or NoneType, optional
            The values ``ritz_vectors.T @ matrix @ guesses`` (which can also
            be computed purely inside the previous Lanczos subspace before the
            collapse, see the code of the :py:`lanczos_iterations` function).
        explicit_symmetrisation : optional
            Explicit symmetrisation to use after orthogonalising the
            subspace vectors. Allows to correct for loss of index or spin
            symmetries during orthogonalisation (type or instance).
        """
        n_problem = matrix.shape[1]   # Problem size

        if not isinstance(guesses, list):
            guesses = [guesses]
        for guess in guesses:
            if not isinstance(guess, AmplitudeVector):
                raise TypeError("One of the guesses is not an AmplitudeVector")
        n_block = len(guesses)  # Lanczos block size

        # For thick restarts, defaults to no restart
        if ritz_values is None:
            n_restart = 0
            ritz_vectors = []  # Y
            ritz_overlaps = np.empty((0, n_block))  # Sigma
            ritz_values = np.empty((0, ))  # Theta
        else:
            n_restart = len(ritz_values)
            if len(ritz_vectors) != n_restart or \
               ritz_overlaps.shape != (n_restart, n_block):
                raise ValueError("Restart vector shape does not agree "
                                 "with problem.")

        self.matrix = matrix
        self.ritz_values = ritz_values
        self.ritz_vectors = ritz_vectors
        self.ritz_overlaps = ritz_overlaps
        self.n_problem = n_problem
        self.n_block = n_block
        self.n_restart = n_restart
        self.ortho = GramSchmidtOrthogonaliser(explicit_symmetrisation)
        self.explicit_symmetrisation = explicit_symmetrisation
        self.timer = Timer()  # TODO More fine-grained timings

        # To be initialised by first call to __next__
        self.lanczos_subspace = []
        self.alphas = []  # Diagonal matrix block of subspace matrix
        self.betas = []   # Side-diagonal matrix blocks of subspace matrix
        self.residual = guesses
        self.n_iter = 0
        self.n_applies = 0

    def __iter__(self):
        return self

    def __next__(self):
        """Advance the iterator, i.e. extend the Lanczos subspace"""
        if self.n_iter == 0:
            # Initialise Lanczos subspace
            v = self.ortho.orthogonalise(self.residual)
            self.lanczos_subspace = v
            r = evaluate(self.matrix @ v)
            alpha = np.empty((self.n_block, self.n_block))
            for p in range(self.n_block):
                alpha[p, :] = v[p] @ r

            # r = r - v * alpha - Y * Sigma
            Sigma, Y = self.ritz_overlaps, self.ritz_vectors
            r = [lincomb(np.hstack(([1], -alpha[:, p], -Sigma[:, p])),
                         [r[p]] + v + Y, evaluate=True)
                 for p in range(self.n_block)]

            # r = r - Y * Y'r (Full reorthogonalisation)
            for p in range(self.n_block):
                r[p] = self.ortho.orthogonalise_against(r[p], self.ritz_vectors)

            self.residual = r
            self.n_iter = 1
            self.n_applies = self.n_block
            self.alphas = [alpha]  # Diagonal matrix block of subspace matrix
            self.betas = []        # Side-diagonal matrix blocks
            return LanczosSubspace(self)

        # Iteration 1 and onwards:
        q = self.lanczos_subspace[-self.n_block:]
        v, beta = self.ortho.qr(self.residual)
        if np.linalg.norm(beta) < np.finfo(float).eps * self.n_problem:
            # No point to go on ... new vectors will be decoupled from old ones
            raise StopIteration()

        # r = A * v - q * beta^T
        self.n_applies += self.n_block
        r = self.matrix @ v
        r = [lincomb(np.hstack(([1], -(beta.T)[:, p])), [r[p]] + q, evaluate=True)
             for p in range(self.n_block)]

        # alpha = v^T * r
        alpha = np.empty((self.n_block, self.n_block))
        for p in range(self.n_block):
            alpha[p, :] = v[p] @ r

        # r = r - v * alpha
        r = [lincomb(np.hstack(([1], -alpha[:, p])), [r[p]] + v, evaluate=True)
             for p in range(self.n_block)]

        # Full reorthogonalisation
        for p in range(self.n_block):
            r[p] = self.ortho.orthogonalise_against(
                r[p], self.lanczos_subspace + self.ritz_vectors
            )

        # Commit results
        self.n_iter += 1
        self.lanczos_subspace.extend(v)
        self.residual = r
        self.alphas.append(alpha)
        self.betas.append(beta)
        return LanczosSubspace(self)


class LanczosSubspace:
    """Container for the Lanczos subspace."""
    def __init__(self, iterator):
        self.__iterator = iterator
        self.n_iter = iterator.n_iter        # Number of iterations
        self.n_restart = iterator.n_restart  # Number of restart vectors
        self.residual = iterator.residual    # Lanczos residual vector(s)
        self.n_block = iterator.n_block      # Block size
        self.n_problem = iterator.n_problem  # Problem size
        self.matrix = iterator.matrix        # Problem matrix
        self.ortho = iterator.ortho          # Orthogonaliser
        self.alphas = iterator.alphas        # Diagonal blocks
        self.betas = iterator.betas          # Side-diagonal blocks
        self.n_applies = iterator.n_applies  # Number of applies

        # Combined set of subspace vectors
        self.subspace = iterator.ritz_vectors + iterator.lanczos_subspace

    @property
    def subspace_matrix(self):
        """
        Return the projection of the problem matrix into the Lanczos subspace
        """
        # TODO Use sparse representation

        n_k = len(self.__iterator.lanczos_subspace)
        n_restart = self.n_restart
        n_ss = n_k + n_restart
        ritz_values = self.__iterator.ritz_values
        ritz_overlaps = self.__iterator.ritz_overlaps

        T = np.zeros((n_ss, n_ss))
        if n_restart > 0:
            T[:n_restart, :n_restart] = np.diag(ritz_values)
            T[:n_restart, n_restart:n_restart + self.n_block] = ritz_overlaps
            T[n_restart:n_restart + self.n_block, :n_restart] = ritz_overlaps.T

        for i, alpha in enumerate(self.alphas):
            rnge = slice(n_restart + i * self.n_block,
                         n_restart + (i + 1) * self.n_block)
            T[rnge, rnge] = alpha
        for i, beta in enumerate(self.betas):
            rnge = slice(n_restart + i * self.n_block,
                         n_restart + (i + 1) * self.n_block)
            rnge_plus = slice(rnge.start + self.n_block, rnge.stop + self.n_block)
            T[rnge, rnge_plus] = beta.T
            T[rnge_plus, rnge] = beta
        return T

    @property
    def rayleigh_extension(self):
        n_k = len(self.__iterator.lanczos_subspace)
        n_ss = n_k + self.n_restart
        b = np.zeros((n_ss, self.n_block))
        for i in range(self.n_block):
            b[n_ss - self.n_block + i, i] = 1
        return b

    @property
    def matrix_product(self):
        """
        Return the reconstructed matrix-vector product for all subspace vectors
        (using the Lanczos relation).
        """
        r, T = self.residual, self.subspace_matrix
        # b = self.rayleigh_extension; V = self.subspace
        # Form AV  = V * T + r * b'
        AV = []
        for i in range(len(self.subspace)):
            # Compute AV[:, i]
            coefficients = []
            vectors = []
            for (j, v) in enumerate(self.subspace):
                if T[j, i] != 0:
                    coefficients.append(T[j, i])
                    vectors.append(v)
            if i >= len(self.subspace) - self.n_block:
                ires = i - (len(self.subspace) - self.n_block)
                coefficients.append(1)
                vectors.append(r[ires])
            AV.append(lincomb(np.array(coefficients), vectors, evaluate=True))
        return AV

    def check_orthogonality(self, tolerance=None):
        if tolerance is None:
            tolerance = self.n_problem * np.finfo(float).eps
        orth = np.array([[SSi @ SSj for SSi in self.subspace]
                         for SSj in self.subspace])
        orth -= np.eye(len(self.subspace))
        orth = np.max(np.abs(orth))
        if orth > tolerance:
            warnings.warn(la.LinAlgWarning(
                "LanczosSubspace has lost orthogonality. "
                "Expect inaccurate results."
            ))
        return orth



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from adcc.timings import Timer


class EigenSolverStateBase:
    def __init__(self, matrix):
        """Initialise an EigenSolverStateBase.

        Parameters
        ----------
        matrix
            Matrix to be diagonalised.
        """
        self.matrix = matrix
        self.eigenvalues = None           # Current eigenvalues
        self.eigenvectors = None          # Current eigenvectors
        self.residual_norms = None        # Current residual norms
        self.converged = False            # Flag whether iteration is converged
        self.n_iter = 0                   # Number of iterations
        self.n_applies = 0                # Number of applies
        self.reortho_triggers = []        # List of reorthogonalisation triggers
        self.timer = Timer()              # Construct a new timer

    def describe(self):
        text = ""

        problem = str(self.matrix)
        algorithm = getattr(self, "algorithm", "")

        if self.converged:
            conv = "converged"
        else:
            conv = "NOT CONVERGED"

        text += "+" + 60 * "-" + "+\n"
        text += "| {0:<41s}  {1:>15s} |\n".format(algorithm, conv)
        text += ("| {0:30s} n_iter={1:<3d}  n_applies={2:<5d} |\n"
                 "".format(problem[:30], self.n_iter, self.n_applies))
        text += ("| n_reortho={0:<7d}  max_overlap_before_reortho={1:<10s}   |\n"
                 "".format(len(self.reortho_triggers),
                           "{:<10.4E}".format(max(self.reortho_triggers))
                           if len(self.reortho_triggers) > 0 else "N/A"))
        text += "+" + 60 * "-" + "+\n"
        text += ("|  #     eigenvalue  res. norm       "
                 "dominant elements       |\n")

        body = "| {0:2d} {1:14.7g}  {2:9.4g}               TODO            |\n"
        for i, vec in enumerate(self.eigenvectors):
            text += body.format(i, self.eigenvalues[i], self.residual_norms[i])
        text += "+" + 60 * "-" + "+"
        return text

    def _repr_pretty_(self, pp, cycle):
        if cycle:
            pp.text("SolverState(...)")
        else:
            pp.text(self.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import warnings
import numpy as np
import inspect
from functools import wraps
from packaging.version import parse

from .timings import Timer


def cached_property(f):
    """
    Decorator for a cached property. From
    https://stackoverflow.com/questions/6428723/python-are-property-fields-being-cached-automatically
    """
    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
            x = self._property_cache[f] = f(self)
            return x
        except KeyError:
            x = self._property_cache[f] = f(self)
            return x

    get.__doc__ = f.__doc__
    # TODO: find more elegant solution for this
    if hasattr(f, "__excitation_property"):
        get.__excitation_property = f.__excitation_property

    return property(get)


def cached_member_function(timer: str = "timer",
                           separate_timings_by_args: bool = True):
    """
    Decorates a member function being called with
    one or more arguments and stores the results
    in field `_function_cache` of the class instance.
    If the class has a timer (defined under the provided name)
    the timings of the (first) call will be measured.

    Parameters
    ----------
    timer: str, optional
        Name of the member variable the :class:`Timer` instance can be found on the
        class instance (default: 'timer'). If the timer is not found no timings
        are measured.
    separate_timings_by_args: bool, optional
        If set the arguments passed to the decorated functions will be included
        in the key under which the timings are stored, i.e., a distinct timer task
        will be generated for each set of arguments. (default: True)
    """
    def decorator(function):
        fname = function.__name__

        # get the function signature and ensure that we don't have any
        # keyword only arguments:
        # func(..., *, kwarg=None, ...) or func(..., **kwargs).
        # If we want to support them we need to add them in a well defined
        # order to the cache key (sort them by name)
        func_signature = inspect.signature(function)
        bad_arg_types = (
            inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD
        )
        if any(arg.kind in bad_arg_types for arg in
               func_signature.parameters.values()):
            raise ValueError("Member functions with keyword only arguments can not "
                             "be wrapped with the cached_member_function.")

        @wraps(function)
        def wrapper(self, *args, **kwargs):
            # convert all arguments to poisitonal arguments and add default
            # arguments for not provided arguments
            bound_args = func_signature.bind(self, *args, **kwargs)
            bound_args.apply_defaults()
            assert not bound_args.kwargs
            args = bound_args.args[1:]  # remove self from args

            try:
                fun_cache = self._function_cache[fname]
            except AttributeError:
                self._function_cache = {}
                fun_cache = self._function_cache[fname] = {}
            except KeyError:
                fun_cache = self._function_cache[fname] = {}

            try:
                return fun_cache[args]
            except KeyError:
                # Record with a timer if possible
                instance_timer = getattr(self, timer, None)
                if isinstance(instance_timer, Timer):
                    timer_task = fname
                    if separate_timings_by_args:
                        descr = '_'.join([str(a) for a in args])
                        timer_task += f"/{descr}"

                    with instance_timer.record(timer_task):
                        # try to evaluate the result if possible
                        result = function(self, *args)
                        if hasattr(result, "evaluate"):
                            result = result.evaluate()
                        fun_cache[args] = result
                else:
                    result = function(self, *args)
                    if hasattr(result, "evaluate"):
                        result = result.evaluate()
                    fun_cache[args] = result
                return result
        return wrapper
    return decorator


def is_module_available(module, min_version=None):
    """Check using importlib if a module is available."""
    import importlib

    try:
        mod = importlib.import_module(module)
    except ImportError:
        return False

    if not min_version:  # No version check
        return True

    if not hasattr(mod, "__version__"):
        warnings.warn(
            f"Could not check module {module} minimal version, "
            "since __version__ tag not found. Proceeding anyway."
        )
        return True

    if parse(mod.__version__) < parse(min_version):
        warnings.warn(
            f"Found module {module}, but its version {mod.__version__} is below "
            f"the least required (== {min_version}). This module will be ignored."
        )
        return False
    return True


def requires_module(name, min_version=None):
    """
    Decorator to check if the module 'name' is available,
    throw ModuleNotFoundError on call if not.
    """
    def inner(function):
        def wrapper(*args, **kwargs):
            fname = function.__name__
            if not is_module_available(name, min_version):
                raise ModuleNotFoundError(
                    f"Function '{fname}' needs module {name}, but it was "
                    f"not found. Solve by running 'pip install {name}' or "
                    f"'conda install {name}' on your system."
                )
            return function(*args, **kwargs)
        wrapper.__doc__ = function.__doc__
        return wrapper
    return inner


def assert_allclose_signfix(actual, desired, atol=0, **kwargs):
    """
    Call assert_allclose, but beforehand normalise the sign
    of the involved arrays (i.e. the two arrays may differ
    up to a sign factor of -1)
    """
    actual, desired = normalise_sign(actual, desired, atol=atol)
    np.testing.assert_allclose(actual, desired, atol=atol, **kwargs)


def normalise_sign(*items, atol=0):
    """
    Normalise the sign of a list of numpy arrays
    """
    def sign(item):
        flat = np.ravel(item)
        flat = flat[np.abs(flat) > atol]
        if flat.size == 0:
            return 1
        else:
            return np.sign(flat[0])
    desired_sign = sign(items[0])
    return tuple(desired_sign / sign(item) * item for item in items)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import libadcc

import opt_einsum

from .AmplitudeVector import AmplitudeVector


def dot(a, b):
    """
    Form the scalar product between two tensors.
    """
    return a.dot(b)


def copy(a):
    """
    Return a copy of the input tensor.
    """
    return a.copy()


def transpose(a, axes=None):
    """
    Return the transpose of a tensor as a *copy*.
    If axes is not given all axes are reversed.
    Else the axes are expect as a tuple of indices,
    e.g. (1,0,2,3) will permute first two axes in the
    returned tensor.
    """
    if axes:
        return a.transpose(axes)
    else:
        return a.transpose()


def empty_like(a):
    """
    Return an empty tensor of the same shape and symmetry as
    the input tensor.
    """
    return a.empty_like()


def zeros_like(a):
    """
    Return a zero tensor of the same shape and symmetry as
    the input tensor.
    """
    return a.zeros_like()


def ones_like(a):
    """
    Return tensor of the same shape and symmetry as
    the input tensor, but initialised to 1,
    that is the canonical blocks are 1 and the
    other ones are symmetry-equivalent (-1 or 0)
    """
    return a.ones_like()


def nosym_like(a):
    """
    Return tensor of the same shape, but without the
    symmetry setup of the input tensor.
    """
    return a.nosym_like()


def lincomb(coefficients, tensors, evaluate=False):
    """
    Form a linear combination from a list of tensors.

    If coefficients is a 1D array, just form a single
    linear combination, else return a list of vectors
    representing the linear combination by reading
    the coefficients row-by-row.

    Parameters
    ----------
    coefficients : list
        Coefficients for the linear combination
    tensors : list
        Tensors for the linear combination
    evaluate : bool
        Should the linear combination be evaluated (True) or should just
        a lazy expression be formed (False). Notice that
        `lincomb(..., evaluate=True)`
        is identical to `lincomb(..., evaluate=False).evaluate()`,
        but the former is generally faster.
    """
    if len(tensors) == 0:
        raise ValueError("List of tensors cannot be empty")
    if len(tensors) != len(coefficients):
        raise ValueError("Number of coefficient values does not match "
                         "number of tensors.")
    if isinstance(tensors[0], AmplitudeVector):
        return AmplitudeVector(**{
            block: lincomb(coefficients, [ten[block] for ten in tensors],
                           evaluate=evaluate)
            for block in tensors[0].blocks
        })
    elif not isinstance(tensors[0], libadcc.Tensor):
        raise TypeError("Tensor type not supported")

    if evaluate:
        # Perform strict evaluation on this linear combination
        return libadcc.linear_combination_strict(coefficients, tensors)
    else:
        # Perform lazy evaluation on this linear combination
        start = float(coefficients[0]) * tensors[0]
        return sum((float(c) * t
                    for (c, t) in zip(coefficients[1:], tensors[1:])), start)


def linear_combination(*args, **kwargs):
    import warnings

    warnings.warn(DeprecationWarning("linear_combination is deprecated and will "
                                     "be removed in 0.17. Use lincomb."))
    return lincomb(*args, **kwargs)


def evaluate(a):
    """Force full evaluation of a tensor expression"""
    if isinstance(a, list):
        return [evaluate(elem) for elem in a]
    elif hasattr(a, "evaluate"):
        return a.evaluate()
    else:
        return libadcc.evaluate(a)


def direct_sum(subscripts, *operands):
    subscripts = subscripts.replace(" ", "")

    def split_signs_symbols(subscripts):
        subscripts = subscripts.replace(",", "+")
        if subscripts[0] not in "+-":
            subscripts = "+" + subscripts
        signs = [x for x in subscripts if x in "+-"]
        symbols = subscripts[1:].replace("-", "+").split("+")
        return signs, symbols

    if "->" in subscripts:
        src, dest = subscripts.split("->")
        signs, src = split_signs_symbols(src)
        # permutation = tuple(dest.index(c) for c in "".join(src))
        permutation = tuple("".join(src).index(c) for c in dest)
    else:
        signs, src = split_signs_symbols(subscripts)
        permutation = None

    if len(src) != len(operands):
        raise ValueError("Number of contraction subscripts does not agree with "
                         "number of operands")
    for i, idcs in enumerate(src):
        if len(idcs) != operands[i].ndim:
            raise ValueError(f"Number of subscripts of {i}-th tensor (== {idcs}) "
                             "does not match dimension of tensor "
                             f"(== {operands[i].ndim}).")

    if signs[0] == "-":
        res = -operands[0]
    else:
        res = operands[0]
    for i, op in enumerate(operands[1:]):
        if signs[i + 1] == "-":
            op = -op
        res = libadcc.direct_sum(res, op)
    if permutation is not None:
        res = res.transpose(permutation)
    return res


def einsum(subscripts, *operands, optimise="auto"):
    """
    Evaluate Einstein summation convention for the operands similar
    to numpy's einsum function. Uses opt_einsum and libadcc to
    perform the contractions.

    Using this function does not evaluate, but returns a contraction
    expression tree where contractions are queued in optimal order.

    Parameters
    ----------

    subscripts : str
        Specifies the subscripts for summation.
    *operands : list of tensors
        These are the arrays for the operation.
    optimise : str, list or bool, optional (default: ``auto``)
        Choose the type of the path optimisation, see
        opt_einsum.contract for details.
    """
    return opt_einsum.contract(subscripts, *operands, optimize=optimise,
                               backend="libadcc")



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import sys
import warnings
import numpy as np

from .SolverStateBase import EigenSolverStateBase
from .explicit_symmetrisation import IndexSymmetrisation

import scipy.linalg as la


class PowerMethodState(EigenSolverStateBase):
    def __init__(self, A):
        super().__init__(A)
        self.residuals = None
        self.algorithm = "power_method"


def default_print(state, identifier, file=sys.stdout):
    """
    A default print function for the power_method callback
    """
    from adcc.timings import strtime, strtime_short

    if identifier == "start" and state.n_iter == 0:
        print("Niter residual  time  Ritz value", file=file)
    elif identifier == "next_iter":
        time_iter = state.timer.current("power_method/iteration")
        fmt = "{n_iter:3d}  {residual:12.5g}  {tstr:5s}"
        print(fmt.format(n_iter=state.n_iter, tstr=strtime_short(time_iter),
                         residual=np.max(state.residual_norms)),
              "", state.eigenvalues[:7], file=file)
    elif identifier == "is_converged":
        soltime = state.timer.total("power_method/iteration")
        print("=== Converged ===", file=file)
        print("    Number of matrix applies:   ", state.n_applies)
        print("    Total solver time:          ", strtime(soltime))


def power_method(A, guess, conv_tol=1e-9, max_iter=70, callback=None,
                 explicit_symmetrisation=IndexSymmetrisation):
    """Use the power iteration to solve for the largest eigenpair of A.

    The power method is a very simple diagonalisation method, which solves
    for the (by magnitude) largest eigenvalue of the matrix `A`.

    Parameters
    ----------
    A
        Matrix object. Only the `@` operator needs to be implemented.
    guess
        Matrix used as a guess
    conv_tol : float
        Convergence tolerance on the l2 norm of residuals to consider
        them converged.
    max_iter : int
        Maximal numer of iterations
    callback
        Callback function called after each iteration
    explicit_symmetrisation
        Explicit symmetrisation to perform during iteration to ensure
        obtaining an eigenvector with matching symmetry criteria.
    """
    if callback is None:
        def callback(state, identifier):
            pass

    if explicit_symmetrisation is not None and \
            isinstance(explicit_symmetrisation, type):
        explicit_symmetrisation = explicit_symmetrisation(A)

    x = guess / np.sqrt(guess @ guess)
    state = PowerMethodState(A)

    def is_converged(state):
        return state.residual_norms[0] < conv_tol

    callback(state, "start")
    state.timer.restart("power_method/iteration")
    for i in range(max_iter):
        state.n_iter += 1
        Ax = A @ x
        state.n_applies += 1

        eigval = x @ (Ax)
        residual = Ax - eigval * x
        residual_norm = np.sqrt(residual @ residual)
        state.eigenvalues = np.array([eigval])
        state.eigenvectors = np.array([x])
        state.residual_norms = np.array([residual_norm])

        callback(state, "next_iter")
        state.timer.restart("power_method/iteration")
        if is_converged(state):
            state.converged = True
            callback(state, "is_converged")
            state.timer.stop("power_method/iteration")
            return state

        if explicit_symmetrisation:
            x = explicit_symmetrisation.symmetrise(Ax)
        else:
            x = Ax
        x = x / np.sqrt(x @ x)

    warnings.warn(la.LinAlgWarning(
        "Power method not converged. Returning intermediate results."))
    return state



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from libadcc import HartreeFockProvider

import psi4

from .EriBuilder import EriBuilder
from ..exceptions import InvalidReference
from ..ElectronicStates import EnergyCorrection
from ..OneParticleOperator import OneParticleOperator


class Psi4OperatorIntegralProvider:
    available: tuple[str, ...] = (
        "overlap", "electric_dipole", "electric_dipole_velocity", "magnetic_dipole",
        "electric_quadrupole", "electric_quadrupole_traceless",
        "pe_induction_elec", "pcm_potential_elec"
    )

    def __init__(self, wfn):
        self.wfn = wfn
        self.backend = "psi4"
        self.mints = psi4.core.MintsHelper(self.wfn)

    @property
    def overlap(self) -> np.ndarray:
        return np.asarray(self.mints.ao_overlap())

    @property
    def electric_dipole(self) -> tuple[np.ndarray, ...]:
        """-sum_i r_i"""
        return tuple(np.asarray(comp) for comp in self.mints.ao_dipole())

    def magnetic_dipole(self, gauge_origin="origin") -> tuple[np.ndarray, ...]:
        """
        The imaginary part of the integral is returned.
        -0.5 * sum_i r_i x p_i
        """
        # TODO: Gauge origin?
        if gauge_origin != (0.0, 0.0, 0.0) and gauge_origin != "origin":
            raise NotImplementedError(
                "For Psi4 only the origin (0.0, 0.0, 0.0) can be selected as"
                " gauge origin for the magnetic dipole operator. "
                f"{gauge_origin} is not valid."
            )
        return tuple(
            0.5 * np.asarray(comp)
            for comp in self.mints.ao_angular_momentum()
        )

    @property
    def electric_dipole_velocity(self) -> tuple[np.ndarray, ...]:
        """
        The imaginary part of the integral is returned.
        -sum_i p_i
        """
        return tuple(-1.0 * np.asarray(comp) for comp in self.mints.ao_nabla())

    def electric_quadrupole(self, gauge_origin="origin") -> tuple[np.ndarray, ...]:
        """-sum_i r_{i, alpha} r_{i, beta}"""
        # TODO: Gauge origin?
        if gauge_origin != (0.0, 0.0, 0.0) and gauge_origin != "origin":
            raise NotImplementedError(
                "For Psi4 only the origin (0.0, 0.0, 0.0) can be selected as"
                " gauge origin for the electric quadrupole operator. "
                f"{gauge_origin} is not valid."
            )
        # Expand 6 upper-triangular components to 9 symmetric matrix entries.
        u = [np.asarray(comp) for comp in self.mints.ao_quadrupole()]
        assert len(u) == 6
        return (u[0], u[1], u[2], u[1], u[3], u[4], u[2], u[4], u[5])

    def electric_quadrupole_traceless(self, gauge_origin="origin"
                                      ) -> tuple[np.ndarray, ...]:
        """
        -0.5 * sum_i (3 * r_{i, alpha} r_{i, beta}
        - delta_{alpha, beta} r_{i}^2)
        """
        # TODO: Gauge origin?
        if gauge_origin != (0.0, 0.0, 0.0) and gauge_origin != "origin":
            raise NotImplementedError(
                "For Psi4 only the origin (0.0, 0.0, 0.0) can be selected as"
                " gauge origin for the traceless electric quadrupole operator. "
                f"{gauge_origin} is not valid."
            )
        # Expand 6 upper-triangular components to 9 symmetric matrix entries.
        u = [np.asarray(comp) for comp in self.mints.ao_traceless_quadrupole()]
        assert len(u) == 6
        return (u[0], u[1], u[2], u[1], u[3], u[4], u[2], u[4], u[5])

    def pe_induction_elec(self, dm: OneParticleOperator) -> psi4.core.Matrix:
        if not hasattr(self.wfn, "pe_state"):
            raise RuntimeError("Can not compute the PE electronic induction "
                               "operator using the given psi4 object.")

        return self.wfn.pe_state.get_pe_contribution(
            psi4.core.Matrix.from_array(dm.to_ndarray()),
            elec_only=True
        )[1]

    def pcm_potential_elec(self, dm: OneParticleOperator) -> psi4.core.Matrix:
        if hasattr(self.wfn, "ddx"):
            return self.wfn.ddx.get_solvation_contributions(
                psi4.core.Matrix.from_array(dm.to_ndarray()),
                elec_only=True, nonequilibrium=True
            )[1]
        elif self.wfn.PCM_enabled():
            return psi4.core.PCM.compute_V(
                self.wfn.get_PCM(),
                psi4.core.Matrix.from_array(dm.to_ndarray())
            )
        raise RuntimeError("Can not compute the electronic PCM potential "
                           "operator using the given psi4 object.")


class Psi4EriBuilder(EriBuilder):
    def __init__(self, wfn, n_orbs, n_orbs_alpha, n_alpha, n_beta, restricted):
        self.wfn = wfn
        self.mints = psi4.core.MintsHelper(self.wfn)
        super().__init__(n_orbs, n_orbs_alpha, n_alpha, n_beta, restricted)

    @property
    def coefficients(self):
        return {
            "Oa": self.wfn.Ca_subset("AO", "OCC"),
            "Ob": self.wfn.Cb_subset("AO", "OCC"),
            "Va": self.wfn.Ca_subset("AO", "VIR"),
            "Vb": self.wfn.Cb_subset("AO", "VIR"),
        }

    def compute_mo_eri(self, blocks, spins):
        coeffs = tuple(self.coefficients[blocks[i] + spins[i]] for i in range(4))
        return np.asarray(self.mints.mo_eri(*coeffs))


class Psi4HFProvider(HartreeFockProvider):
    """
        This implementation is only valid
        if no orbital reordering is required.
    """
    def __init__(self, wfn):
        # Do not forget the next line,
        # otherwise weird errors result
        super().__init__()
        self.wfn = wfn
        self.eri_builder = Psi4EriBuilder(self.wfn, self.n_orbs, self.wfn.nmo(),
                                          wfn.nalpha(), wfn.nbeta(),
                                          self.restricted)
        self.operator_integral_provider = Psi4OperatorIntegralProvider(self.wfn)

        self.environment = None
        self.environment_implementation = None
        if hasattr(self.wfn, "pe_state"):
            self.environment = "pe"
            self.environment_implementation = "cppe"
        elif hasattr(self.wfn, "ddx"):
            self.environment = "pcm"
            self.environment_implementation = "ddx"
        elif self.wfn.PCM_enabled():
            self.environment = "pcm"
            self.environment_implementation = "pcmsolver"

    def pe_energy(self, dm, elec_only=True):
        density_psi = psi4.core.Matrix.from_array(dm.to_ndarray())
        e_pe, _ = self.wfn.pe_state.get_pe_contribution(density_psi,
                                                        elec_only=elec_only)
        return e_pe

    def pcm_energy(self, dm, elec_only=True):
        psi_dm = psi4.core.Matrix.from_array(dm.to_ndarray())
        # computes the Fock matrix contribution.
        # By contraction with the tdm, the electronic energy contribution is
        # calculated.
        if self.environment_implementation == "ddx":
            V_pcm = self.wfn.ddx.get_solvation_contributions(
                psi_dm, elec_only=elec_only, nonequilibrium=True
            )[1]
        elif self.environment_implementation == "pcmsolver":
            V_pcm = psi4.core.PCM.compute_V(self.wfn.get_PCM(), psi_dm)
        return np.einsum("uv,uv->", dm.to_ndarray(), V_pcm.to_array())

    @property
    def excitation_energy_corrections(self):
        ret = []
        if self.environment == "pe":
            ptlr = EnergyCorrection(
                "pe_ptlr_correction",
                lambda view: 2.0 * self.pe_energy(view.transition_dm_ao,
                                                  elec_only=True)
            )
            ptss = EnergyCorrection(
                "pe_ptss_correction",
                lambda view: self.pe_energy(view.state_diffdm_ao,
                                            elec_only=True)
            )
            ret.extend([ptlr, ptss])
        if self.environment == "pcm":
            ptlr = EnergyCorrection(
                "pcm_ptlr_correction",
                lambda view: self.pcm_energy(view.transition_dm_ao,
                                             elec_only=True)
            )
            ret.extend([ptlr])
        return {ec.name: ec for ec in ret}

    def get_backend(self):
        return "psi4"

    def get_conv_tol(self):
        conv_tol = psi4.core.get_option("SCF", "E_CONVERGENCE")
        # RMS value of the orbital gradient
        conv_tol_grad = psi4.core.get_option("SCF", "D_CONVERGENCE")
        threshold = max(conv_tol, conv_tol_grad)
        return threshold

    def get_restricted(self):
        return isinstance(self.wfn, (psi4.core.RHF, psi4.core.ROHF))

    def get_energy_scf(self):
        return self.wfn.energy()

    def get_spin_multiplicity(self):
        return self.wfn.molecule().multiplicity()

    def get_n_orbs_alpha(self):
        return self.wfn.nmo()

    def get_n_bas(self):
        return self.wfn.basisset().nbf()

    def get_nuclear_multipole(self, order, gauge_origin=(0, 0, 0)):
        molecule = self.wfn.molecule()
        if order == 0:
            # The function interface needs to be a np.array on return
            return np.array([sum(molecule.charge(i)
                                 for i in range(molecule.natom()))])
        elif order == 1:
            dip_nuclear = molecule.nuclear_dipole()
            return np.array([dip_nuclear[0], dip_nuclear[1], dip_nuclear[2]])
        else:
            raise NotImplementedError("get_nuclear_multipole with order > 1")

    def transform_gauge_origin_to_xyz(self, gauge_origin):
        raise NotImplementedError("transform_gauge_origin_to_xyz not implemented.")

    def fill_orbcoeff_fb(self, out):
        mo_coeff_a = np.asarray(self.wfn.Ca())
        mo_coeff_b = np.asarray(self.wfn.Cb())
        mo_coeff = (mo_coeff_a, mo_coeff_b)
        out[:] = np.transpose(
            np.hstack((mo_coeff[0], mo_coeff[1]))
        )

    def fill_occupation_f(self, out):
        out[:] = np.hstack((
            np.asarray(self.wfn.occupation_a()),
            np.asarray(self.wfn.occupation_b())
        ))

    def fill_orben_f(self, out):
        orben_a = np.asarray(self.wfn.epsilon_a())
        orben_b = np.asarray(self.wfn.epsilon_b())
        out[:] = np.hstack((orben_a, orben_b))

    def fill_fock_ff(self, slices, out):
        diagonal = np.empty(self.n_orbs)
        self.fill_orben_f(diagonal)
        out[:] = np.diag(diagonal)[slices]

    def fill_eri_ffff(self, slices, out):
        self.eri_builder.fill_slice_symm(slices, out)

    def fill_eri_phys_asym_ffff(self, slices, out):
        raise NotImplementedError("fill_eri_phys_asym_ffff not implemented.")

    def has_eri_phys_asym_ffff(self):
        return False

    def flush_cache(self):
        self.eri_builder.flush_cache()


def import_scf(wfn):
    if not isinstance(wfn, psi4.core.HF):
        raise InvalidReference(
            "Only psi4.core.HF and its subtypes are supported references in "
            "backends.psi4.import_scf. This indicates that you passed an "
            "unsupported SCF reference. Make sure you did a restricted or "
            "unrestricted HF calculation."
        )

    if not isinstance(wfn, (psi4.core.RHF, psi4.core.UHF)):
        raise InvalidReference("Right now only RHF and UHF references are "
                               "supported for Psi4.")

    # TODO This is not fully correct, because the core.Wavefunction object
    #      has an internal, but py-invisible Options structure, which contains
    #      the actual set of options ... theoretically they could differ
    scf_type = psi4.core.get_global_option('SCF_TYPE')
    # CD = Choleski, DF = density-fitting
    unsupported_scf_types = ["CD", "DISK_DF", "MEM_DF"]
    if scf_type in unsupported_scf_types:
        raise InvalidReference("Unsupported Psi4 SCF_TYPE, should not be one "
                               f"of {unsupported_scf_types}")

    if wfn.nirrep() > 1:
        raise InvalidReference("The passed Psi4 wave function object needs to "
                               "have exactly one irrep, i.e. be of C1 symmetry.")

    # Psi4 throws an exception if SCF is not converged, so there is no need
    # to assert that here.
    provider = Psi4HFProvider(wfn)
    return provider


def run_hf(xyz, basis, charge=0, multiplicity=1, conv_tol=1e-11,
           conv_tol_grad=1e-9, max_iter=150, pe_options=None):
    basissets = {
        "sto3g": "sto-3g",
        "def2tzvp": "def2-tzvp",
        "ccpvdz": "cc-pvdz",
    }

    # needed for PE and PCM tests
    psi4.core.clean_options()
    mol = psi4.geometry(f"""
        {charge} {multiplicity}
        {xyz}
        symmetry c1
        units au
        no_reorient
        no_com
    """)

    psi4.core.be_quiet()
    psi4.set_options({
        'basis': basissets.get(basis, basis),
        'scf_type': 'pk',
        'e_convergence': conv_tol,
        'd_convergence': conv_tol_grad,
        'maxiter': max_iter,
        'reference': "RHF",
    })
    if pe_options:
        psi4.set_options({"pe": "true"})
        psi4.set_module_options("pe", {"potfile": pe_options["potfile"]})

    if multiplicity != 1:
        psi4.set_options({
            'reference': "UHF",
            'maxiter': max_iter + 500,
            'soscf': 'true'
        })

    _, wfn = psi4.energy('SCF', return_wfn=True, molecule=mol)
    psi4.core.clean()
    return wfn



from . import davidson
from .SolverStateBase import EigenSolverStateBase
from .explicit_symmetrisation import (IndexSpinSymmetrisation,
                                      IndexSymmetrisation)

__all__ = ["IndexSymmetrisation", "IndexSpinSymmetrisation",
           "davidson", "EigenSolverStateBase"]
